from cvxopt import matrix, mul, spmatrix, div, sin, cos
from .dcbase import DCBase
from ..utils.math import zeros, conj, polar

from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA
import logging

logger = logging.getLogger(__name__)


class VSC(DCBase):
    """VSC model for power flow study"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'AC/DC'
        self._name = 'VSC'
        self._ac = {'bus': ['a', 'v']}
        self._params.extend([
            'rsh', 'xsh', 'vshmax', 'vshmin', 'Ishmax', 'pref0', 'qref0',
            'vref0', 'v0', 'vdcref0', 'k0', 'k1', 'k2', 'droop', 'K', 'vhigh',
            'vlow'
        ])
        self._data.update({
            'rsh': 0.0025,
            'xsh': 0.06,
            'vshmax': 1.1,
            'vshmin': 0.9,
            'Ishmax': 2,
            'bus': None,
            'control': None,
            'v0': 1.0,
            'pref0': 0,
            'qref0': 0,
            'vref0': 1.0,
            'vdcref0': 1.0,
            'k0': 0,
            'k1': 0,
            'k2': 0,
            'droop': False,
            'K': 0,
            'vhigh': 9999,
            'vlow': 0.0,
        })
        self._units.update({
            'rsh': 'ohm',
            'xsh': 'ohm',
            'vshmax': 'pu',
            'vshmin': 'pu',
            'Ishmax': 'pu',
            'vref0': 'pu',
            'pref0': 'pu',
            'qref0': 'pu',
            'vdcref0': 'pu',
            'droop': 'boolean',
            'vhigh': 'pu',
            'vlow': 'pu',
        })
        self._descr.update({
            'rsh':
            'AC interface resistance',
            'xsh':
            'AC interface reactance',
            'vshmax':
            'Maximum ac interface voltage',
            'vshmin':
            'Minimum ac interface voltage',
            'Ishmax':
            'Maximum ac current',
            'vref0':
            'AC voltage setting',
            'pref0':
            'AC active power setting',
            'qref0':
            'AC reactive power setting',
            'vdcref0':
            'DC voltage setting',
            'k0':
            'Loss coefficient - constant',
            'k1':
            'Loss coefficient - linear',
            'k2':
            'Loss coefficient - quadratic',
            'droop':
            'Enable dc voltage droop control',
            'K':
            'Droop coefficient',
            'vhigh':
            'Upper voltage threshold in droop control',
            'vlow':
            'Lower voltage threshold in droop control',
            'control':
            'Control method of this VSC in PQ, PV, vQ or vV',
            'bus':
            'AC bus idx',
            'v0':
            'AC voltage initial guess for PV and PQ controlled VSC'
        })
        self._algebs.extend(['ash', 'vsh', 'psh', 'qsh', 'pdc', 'Ish'])
        self._unamey.extend(['ash', 'vsh', 'psh', 'qsh', 'pdc', 'Ish'])
        self._fnamey.extend(
            ['\\theta_{sh}', 'V_{sh}', 'P_{sh}', 'Q_{sh}', 'P_{dc}', 'I_{sh}'])
        self._mandatory.extend(['bus', 'control'])
        self._service.extend([
            'Zsh', 'Ysh', 'glim', 'ylim', 'vdcref', 'R', 'PQ', 'PV', 'vV', 'vQ'
        ])
        self.calls.update({
            'init0': True,
            'pflow': True,
            'gcall': True,
            'gycall': True,
            'jac0': True,
            'shunt': True,
        })
        self._init()
        self.glim = []
        self.ylim = []
        self.vio = {}

    def setup(self):
        super().setup()
        self.K = mul(self.K, self.droop)
        self.R = matrix(0, (self.n, 1), 'd')
        self.Zsh = self.rsh + 1j * self.xsh
        self.Ysh = div(1, self.Zsh)

        self.PQ = zeros(self.n, 1)
        self.PV = zeros(self.n, 1)
        self.vV = zeros(self.n, 1)
        self.vQ = zeros(self.n, 1)
        for idx, cc in enumerate(self.control):
            if cc not in ['PQ', 'PV', 'vV', 'vQ']:
                raise KeyError(
                    'VSC {0} control parameter {1} is invalid.'.format(
                        self.name[idx], cc))
            self.__dict__[cc][idx] = 1

    def init0(self, dae):
        # behind-transformer AC theta_sh and V_sh - must assign first
        dae.y[self.ash] = dae.y[self.a] + 1e-6
        dae.y[self.vsh] = mul(self.v0, 1 - self.vV) + mul(self.vref0,
                                                          self.vV) + 1e-6

        Vm = polar(dae.y[self.v], dae.y[self.a])
        Vsh = polar(dae.y[self.vsh], dae.y[self.ash])  # Initial value for Vsh
        IshC = conj(div(Vsh - Vm, self.Zsh))

        # PQ PV and V control initials on converters
        dae.y[self.psh] = mul(self.pref0, self.PQ + self.PV)
        dae.y[self.qsh] = mul(self.qref0, self.PQ)
        dae.y[self.v1] = dae.y[self.v2] + mul(
            dae.y[self.v1], 1 - self.vV) + mul(self.vdcref0, self.vV)

        # PV and V control on AC buses
        dae.y[self.v] = mul(dae.y[self.v], 1 - self.PV - self.vV) + mul(
            self.vref0, self.PV + self.vV)

        # Converter current initial
        dae.y[self.Ish] = abs(IshC)

        # Converter dc power output
        dae.y[self.pdc] = mul(Vsh, IshC).real() + \
            (self.k0 + mul(self.k1, dae.y[self.Ish]) + mul(
                self.k2, mul(dae.y[self.Ish], dae.y[self.Ish])))

    def gcall(self, dae):
        if sum(self.u) == 0:
            return

        Vm = polar(dae.y[self.v], dae.y[self.a])
        Vsh = polar(dae.y[self.vsh], dae.y[self.ash])
        Ish = mul(self.Ysh, Vsh - Vm)
        IshC = conj(Ish)
        Ssh = mul(Vm, IshC)

        # check the Vsh and Ish limits during PF iterations
        vupper = list(abs(Vsh) - self.vshmax)
        vlower = list(abs(Vsh) - self.vshmin)
        iupper = list(abs(IshC) - self.Ishmax)
        # check for Vsh and Ish limit violations
        if self.system.pflow.niter >= self.system.pflow.config.ipv2pq:
            for i in range(self.n):
                if self.u[i] and (vupper[i] > 0 or vlower[i] < 0
                                  or iupper[i] > 0):
                    if i not in self.vio.keys():
                        self.vio[i] = list()
                    if vupper[i] > 0:
                        if 'vmax' not in self.vio[i]:
                            self.vio[i].append('vmax')
                            logger.debug(
                                ' * Vmax reached for VSC_{0}'.format(i))
                    elif vlower[i] < 0:
                        if 'vmin' not in self.vio[i]:
                            self.vio[i].append('vmin')
                            logger.debug(
                                ' * Vmin reached for VSC_{0}'.format(i))
                    if iupper[i] > 0:
                        if 'Imax' not in self.vio[i]:
                            self.vio[i].append('Imax')
                            logger.debug(
                                ' * Imax reached for VSC_{0}'.format(i))

        # AC interfaces - power
        dae.g -= spmatrix(
            mul(self.u, dae.y[self.psh]), self.a, [0] * self.n, (dae.m, 1),
            'd')
        dae.g -= spmatrix(
            mul(self.u, dae.y[self.qsh]), self.v, [0] * self.n, (dae.m, 1),
            'd')

        # DC interfaces - current
        above = list(dae.y[self.v1] - self.vhigh)
        below = list(dae.y[self.v1] - self.vlow)
        above = matrix([1 if i > 0 else 0 for i in above])
        below = matrix([1 if i < 0 else 0 for i in below])
        self.R = mul(above or below, self.K)
        self.vdcref = mul(self.droop, above, self.vhigh) + mul(
            self.droop, below, self.vlow)
        dae.g += spmatrix(
            div(mul(self.u, dae.y[self.pdc]), dae.y[self.v1] - dae.y[self.v2]),
            self.v1, [0] * self.n, (dae.m, 1), 'd')
        dae.g -= spmatrix(
            div(mul(self.u, dae.y[self.pdc]),
                dae.y[self.v1] - dae.y[self.v2]), self.v2, [0] * self.n,
            (dae.m, 1), 'd')  # negative current injection

        dae.g[self.ash] = mul(self.u, Ssh.real() - dae.y[self.psh])  # (2)
        dae.g[self.vsh] = mul(self.u, Ssh.imag() - dae.y[self.qsh])  # (3)

        # PQ, PV or V control
        dae.g[self.psh] = mul(
            dae.y[self.psh] + mul(self.R, dae.y[self.v1] - self.vdcref) -
            self.pref0, self.PQ + self.PV, self.u) + mul(
                (dae.y[self.v1] - dae.y[self.v2]) - self.vdcref0,
                self.vV + self.vQ, self.u)  # (12), (15)
        dae.g[self.qsh] = mul(
            dae.y[self.qsh] - self.qref0, self.PQ + self.vQ, self.u) + mul(
                dae.y[self.v] - self.vref0, self.PV + self.vV,
                self.u)  # (13), (16)

        for comp, var in self.vio.items():
            for count, item in enumerate(var):
                if item == 'vmax':
                    yidx = self.vsh[comp]
                    ylim = self.vshmax[comp]
                elif item == 'vmin':
                    yidx = self.vsh[comp]
                    ylim = self.vshmin[comp]
                elif item == 'Imax':
                    yidx = self.Ish[comp]
                    ylim = self.Ishmax[comp]
                else:
                    raise NameError(
                        'Unknown limit variable name <{0}>.'.format(item))

                if count == 0:
                    idx = self.qsh[comp]
                    self.switch(comp, 'Q')
                else:
                    idx = self.psh[comp]
                    self.switch(comp, 'P')

                self.system.dae.factorize = True

                dae.g[idx] = dae.y[yidx] - ylim
                if idx not in self.glim:
                    self.glim.append(idx)
                if yidx not in self.ylim:
                    self.ylim.append(yidx)

        dae.g[self.Ish] = mul(self.u, abs(IshC) - dae.y[self.Ish])  # (10)

        dae.g[self.pdc] = mul(
            self.u,
            mul(Vsh, IshC).real() - dae.y[self.pdc] + (self.k0 + mul(
                self.k1, dae.y[self.Ish]) + mul(self.k2, dae.y[self.Ish]**2)))

    def switch(self, idx, control):
        """Switch a single control of <idx>"""
        old = None
        new = None
        if control == 'Q':
            if self.PQ[idx] == 1:
                old = 'PQ'
                new = 'PV'
            elif self.vQ[idx] == 1:
                old = 'vQ'
                new = 'vV'
        elif control == 'P':
            if self.PQ[idx] == 1:
                old = 'PQ'
                new = 'vQ'
            elif self.PV[idx] == 1:
                old = 'PV'
                new = 'vV'
        elif control == 'V':
            if self.PV[idx] == 1:
                old = 'PV'
                new = 'PQ'
            elif self.vV[idx] == 1:
                old = 'vV'
                new = 'vQ'
        elif control == 'v':
            if self.vQ[idx] == 1:
                old = 'vQ'
                new = 'PQ'
            elif self.vV[idx] == 1:
                old = 'vV'
                new = 'PV'
        if old and new:
            self.__dict__[old][idx] = 0
            self.__dict__[new][idx] = 1

    def gycall(self, dae):
        if sum(self.u) == 0:
            return
        Zsh = self.rsh + 1j * self.xsh
        iZsh = div(self.u, abs(Zsh))
        Vh = polar(dae.y[self.v], dae.y[self.a] * 1j)
        Vsh = polar(dae.y[self.vsh], dae.y[self.ash] * 1j)
        Ish = div(Vsh - Vh + 1e-6, Zsh)
        iIsh = div(self.u, Ish)

        gsh = div(self.u, Zsh).real()
        bsh = div(self.u, Zsh).imag()
        V = dae.y[self.v]
        theta = dae.y[self.a]
        Vsh = dae.y[self.vsh]
        thetash = dae.y[self.ash]
        Vdc = dae.y[self.v1] - dae.y[self.v2]
        iVdc2 = div(self.u, Vdc**2)

        dae.add_jac(Gy, div(self.u, Vdc), self.v1, self.pdc)
        dae.add_jac(Gy, -mul(self.u, dae.y[self.pdc], iVdc2), self.v1, self.v1)
        dae.add_jac(Gy, mul(self.u, dae.y[self.pdc], iVdc2), self.v1, self.v2)

        dae.add_jac(Gy, -div(self.u, Vdc), self.v2, self.pdc)
        dae.add_jac(Gy, mul(self.u, dae.y[self.pdc], iVdc2), self.v2, self.v1)
        dae.add_jac(Gy, -mul(self.u, dae.y[self.pdc], iVdc2), self.v2, self.v2)

        dae.add_jac(
            Gy, -2 * mul(gsh, V) + mul(gsh, Vsh, cos(theta - thetash)) + mul(
                bsh, Vsh, sin(theta - thetash)), self.ash, self.v)
        dae.add_jac(
            Gy,
            mul(gsh, V, cos(theta - thetash)) + mul(
                bsh, V, sin(theta - thetash)), self.ash, self.vsh)
        dae.add_jac(
            Gy, -mul(gsh, V, Vsh, sin(theta - thetash)) + mul(
                bsh, V, Vsh, cos(theta - thetash)), self.ash, self.a)
        dae.add_jac(
            Gy,
            mul(gsh, V, Vsh, sin(theta - thetash)) - mul(
                bsh, V, Vsh, cos(theta - thetash)), self.ash, self.ash)

        dae.add_jac(
            Gy, 2 * mul(bsh, V) + mul(gsh, Vsh, sin(theta - thetash)) - mul(
                bsh, Vsh, cos(theta - thetash)), self.vsh, self.v)
        dae.add_jac(
            Gy,
            mul(gsh, V, sin(theta - thetash)) - mul(
                bsh, V, cos(theta - thetash)), self.vsh, self.vsh)
        dae.add_jac(
            Gy,
            mul(gsh, V, Vsh, cos(theta - thetash)) + mul(
                bsh, V, Vsh, sin(theta - thetash)), self.vsh, self.a)
        dae.add_jac(
            Gy, -mul(gsh, V, Vsh, cos(theta - thetash)) - mul(
                bsh, V, Vsh, sin(theta - thetash)), self.vsh, self.ash)

        dae.add_jac(
            Gy, 0.5 * mul(self.u, 2 * V - 2 * mul(Vsh, cos(theta - thetash)),
                          abs(iIsh),
                          abs(iZsh)**2), self.Ish, self.v)
        dae.add_jac(
            Gy, 0.5 * mul(self.u, 2 * Vsh - 2 * mul(V, cos(theta - thetash)),
                          abs(iIsh),
                          abs(iZsh)**2), self.Ish, self.vsh)
        dae.add_jac(
            Gy, 0.5 * mul(self.u, 2 * V, Vsh, sin(theta - thetash), abs(iIsh),
                          abs(iZsh)**2), self.Ish, self.a)
        dae.add_jac(
            Gy, 0.5 * mul(self.u, 2 * V, Vsh, -sin(theta - thetash), abs(iIsh),
                          abs(iZsh)**2), self.Ish, self.ash)

        dae.add_jac(Gy, -2 * mul(self.u, self.k2, dae.y[self.Ish]), self.pdc,
                    self.Ish)

        dae.add_jac(
            Gy,
            mul(2 * gsh, Vsh) - mul(gsh, V, cos(theta - thetash)) + mul(
                bsh, V, sin(theta - thetash)), self.pdc, self.vsh)
        dae.add_jac(
            Gy, -mul(gsh, Vsh, cos(theta - thetash)) + mul(
                bsh, Vsh, sin(theta - thetash)), self.pdc, self.v)
        dae.add_jac(
            Gy,
            mul(gsh, V, Vsh, sin(theta - thetash)) + mul(
                bsh, V, Vsh, cos(theta - thetash)), self.pdc, self.a)
        dae.add_jac(
            Gy, -mul(gsh, V, Vsh, sin(theta - thetash)) - mul(
                bsh, V, Vsh, cos(theta - thetash)), self.pdc, self.ash)

        for gidx, yidx in zip(self.glim, self.ylim):
            dae.set_jac(Gy, 0.0, [gidx] * dae.m, range(dae.m))
            dae.set_jac(Gy, 1.0, [gidx], [yidx])
            dae.set_jac(Gy, -1e-6, [gidx], [gidx])

    def jac0(self, dae):
        dae.add_jac(Gy0, -1e-6, self.v1, self.v1)
        dae.add_jac(Gy0, -1e-6, self.v2, self.v2)
        dae.add_jac(Gy0, -1e-6, self.ash, self.ash)
        dae.add_jac(Gy0, -1e-6, self.vsh, self.vsh)

        dae.add_jac(Gy0, -self.u, self.ash, self.psh)
        dae.add_jac(Gy0, -self.u, self.vsh, self.qsh)

        dae.add_jac(Gy0,
                    mul(self.u, self.PQ + self.PV) - 1e-6, self.psh, self.psh)
        dae.add_jac(Gy0, mul(self.u, self.vV), self.psh, self.v1)
        dae.add_jac(Gy0, -mul(self.u, self.vV), self.psh, self.v2)
        dae.add_jac(Gy0, mul(self.PV + self.PQ, self.u, self.R), self.psh,
                    self.v1)

        dae.add_jac(Gy0, mul(self.u, self.PQ) - 1e-6, self.qsh, self.qsh)
        dae.add_jac(Gy0, mul(self.u, self.PV + self.vV), self.qsh, self.v)

        dae.add_jac(Gy0, -self.u, self.a, self.psh)
        dae.add_jac(Gy0, -self.u, self.v, self.qsh)

        dae.add_jac(Gy0, -self.u - 1e-6, self.Ish, self.Ish)

        dae.add_jac(Gy0, -self.u - 1e-6, self.pdc, self.pdc)
        dae.add_jac(Gy0, mul(self.u, self.k1), self.pdc, self.Ish)

    def disable(self, idx):
        """Disable an element and reset the outputs"""
        if idx not in self.uid.keys():
            self.log('Element index {0} does not exist.'.format(idx))
            return
        self.u[self.uid[idx]] = 0


class VSC1_Common(DCBase):
    """Common equations for VSC1"""

    def __init__(self, system, name):
        super(VSC1_Common, self).__init__(system, name)
        self._group = 'AC/DC'
        self._data.update({
            'vsc': None,
            'Kdc': 0,
            'Kf': 0,
            'KV': 0,
            'Kp': 0,
        })
        self._mandatory.extend(['vsc'])
        self._descr.update({
            'vsc': 'static vsc idx',
            'Kdc': 'droop of dc voltage on active power',
            'Kf': 'droop of ac frequency on power',
            'KV': 'droop of ac voltage on reactive power',
            'Kp': 'droop of power on dc voltage',
        })
        self._params.extend(['Kdc', 'Kf', 'KV', 'Kp'])
        self._algebs.extend(['ref1', 'ref2', 'vd', 'vq', 'p', 'q'])
        self._fnamey.extend(['ref_1', 'ref2', 'v_d', 'v_q', 'P', 'Q'])
        self._service.extend(['ref10', 'ref20'])

        self.calls.update({
            'fcall': True,
            'fxcall': True,
            'gcall': True,
            'gycall': True,
            'jac0': True,
            'init1': True
        })
        self._mandatory.remove('node1')
        self._mandatory.remove('node2')
        self._mandatory.remove('Vdcn')
        self._ac = {}
        self._dc = {}

    def servcall(self, dae):
        self.copy_data_ext('VSC', 'Sn', 'Sn', self.vsc)
        self.copy_data_ext('VSC', 'Vn', 'Vn', self.vsc)
        self.copy_data_ext('VSC', 'Vdcn', 'Vdcn', self.vsc)
        self.copy_data_ext('VSC', 'rsh', 'rsh', self.vsc)
        self.copy_data_ext('VSC', 'xsh', 'xsh', self.vsc)
        self.copy_data_ext('VSC', 'PQ', 'PQ', self.vsc)
        self.copy_data_ext('VSC', 'PV', 'PV', self.vsc)
        self.copy_data_ext('VSC', 'vQ', 'vQ', self.vsc)
        self.copy_data_ext('VSC', 'vV', 'vV', self.vsc)
        self.copy_data_ext('VSC', 'a', 'a', self.vsc)
        self.copy_data_ext('VSC', 'v', 'v', self.vsc)
        self.copy_data_ext('VSC', 'v1', 'v1', self.vsc)
        self.copy_data_ext('VSC', 'v2', 'v2', self.vsc)
        self.copy_data_ext('VSC', 'psh', 'pref0', self.vsc)
        self.copy_data_ext('VSC', 'qsh', 'qref0', self.vsc)
        self.copy_data_ext('VSC', 'bus', 'bus', self.vsc)
        self.copy_data_ext(
            'BusFreq', 'w', 'w',
            self.bus)  # TODO: BusFreq idx must be the same as bus idx

        self.pref0 = dae.y[self.pref0]
        self.qref0 = dae.y[self.qref0]
        self.ref10 = mul(self.pref0, self.PQ + self.PV) + mul(
            dae.y[self.v1] - dae.y[self.v2], self.vQ + self.vV)
        self.ref20 = mul(self.qref0, self.PQ + self.vQ) + mul(
            dae.y[self.v], self.PV + self.vV)
        self.v120 = matrix(self.v12)
        self.V0 = matrix(dae.y[self.v])

    def init1(self, dae):
        self.servcall(dae)

        self.pll_init1(dae)

        dae.y[self.ref1] = self.ref10
        dae.y[self.ref2] = self.ref20
        dae.y[self.vd] = mul(dae.y[self.v],
                             cos(dae.y[self.a] - dae.y[self.adq]))
        dae.y[self.vq] = -mul(dae.y[self.v],
                              sin(dae.y[self.a] - dae.y[self.adq]))

        self.current_init1(dae)
        self.outer_init1(dae)

        dae.y[self.p] = mul(dae.x[self.Id], dae.y[self.vd]) + mul(
            dae.x[self.Iq], dae.y[self.vq])
        dae.y[self.q] = mul(dae.x[self.Iq], dae.y[self.vd]) - mul(
            dae.x[self.Id], dae.y[self.vq])

        self.power_init1(dae)
        for idx in self.vsc:
            self.system.VSC.disable(idx)

    def gcall(self, dae):
        self.pll_gcall(dae)
        self.outer_gcall(dae)
        self.current_gcall(dae)
        self.power_gcall(dae)

        dae.g[self.ref1] += self.ref10 - dae.y[self.ref1] \
            + mul(self.PQ + self.PV, mul(self.v12 - self.v120, self.Kdc)) \
            - mul(self.PQ + self.PV, mul(dae.x[self.w] - 1), self.Kf) \
            + mul(self.vV + self.vQ, self.Kp, dae.y[self.p] - self.pref0)\
            - mul(self.vV + self.vQ, mul(dae.x[self.w] - 1), self.Kf)

        dae.g[self.ref2] += self.ref20 - dae.y[self.ref2]
        dae.g[self.vd] = -dae.y[self.vd] + mul(
            dae.y[self.v], cos(dae.y[self.a] - dae.y[self.adq]))
        dae.g[self.vq] = -dae.y[self.vq] - mul(
            dae.y[self.v], sin(dae.y[self.a] - dae.y[self.adq]))
        dae.g[self.p] = -dae.y[self.p] + mul(
            dae.x[self.Id], dae.y[self.vd]) + mul(dae.x[self.Iq],
                                                  dae.y[self.vq])
        dae.g[self.q] = -dae.y[self.q] + mul(
            dae.x[self.Iq], dae.y[self.vd]) - mul(dae.x[self.Id],
                                                  dae.y[self.vq])
        dae.g += spmatrix(-dae.y[self.p], self.a, [0] * self.n, (dae.m, 1),
                          'd')
        dae.g += spmatrix(-dae.y[self.q], self.v, [0] * self.n, (dae.m, 1),
                          'd')

    def fcall(self, dae):
        self.pll_fcall(dae)
        self.outer_fcall(dae)
        self.current_fcall(dae)
        self.power_fcall(dae)

    def gycall(self, dae):
        self.pll_gycall(dae)
        self.outer_gycall(dae)
        self.current_gycall(dae)
        self.power_gycall(dae)

        dae.add_jac(Gy, cos(dae.y[self.a] - dae.y[self.adq]), self.vd, self.v)
        dae.add_jac(Gy,
                    -mul(dae.y[self.v], sin(dae.y[self.a] - dae.y[self.adq])),
                    self.vd, self.a)
        dae.add_jac(Gy, mul(dae.y[self.v],
                            sin(dae.y[self.a] - dae.y[self.adq])), self.vd,
                    self.adq)
        dae.add_jac(Gy, -sin(dae.y[self.a] - dae.y[self.adq]), self.vq, self.v)
        dae.add_jac(Gy,
                    -mul(dae.y[self.v], cos(dae.y[self.a] - dae.y[self.adq])),
                    self.vq, self.a)
        dae.add_jac(Gy, mul(dae.y[self.v],
                            cos(dae.y[self.a] - dae.y[self.adq])), self.vq,
                    self.adq)
        dae.add_jac(Gy, dae.x[self.Iq], self.p, self.vq)
        dae.add_jac(Gy, dae.x[self.Id], self.p, self.vd)
        dae.add_jac(Gy, -dae.x[self.Id], self.q, self.vq)
        dae.add_jac(Gy, dae.x[self.Iq], self.q, self.vd)

    def fxcall(self, dae):
        self.pll_fxcall(dae)
        self.outer_fxcall(dae)
        self.current_fxcall(dae)
        self.power_fxcall(dae)

        dae.add_jac(Gx, dae.y[self.vd], self.p, self.Id)
        dae.add_jac(Gx, dae.y[self.vq], self.p, self.Iq)
        dae.add_jac(Gx, -dae.y[self.vq], self.q, self.Id)
        dae.add_jac(Gx, dae.y[self.vd], self.q, self.Iq)

    def jac0(self, dae):
        self.pll_jac0(dae)
        self.outer_jac0(dae)
        self.current_jac0(dae)
        self.power_jac0(dae)

        dae.add_jac(Gy0, mul(self.PQ + self.PV, self.Kdc), self.ref1, self.v1)
        dae.add_jac(Gy0, -mul(self.PQ + self.PV, self.Kdc), self.ref1, self.v2)
        dae.add_jac(Gx0, -mul(self.PQ + self.PV, self.Kf), self.ref1, self.w)
        dae.add_jac(Gy0, mul(self.vV + self.vQ, self.Kp), self.ref1, self.p)
        dae.add_jac(Gx0, -mul(self.vV + self.vQ, self.Kf), self.ref1, self.w)

        dae.add_jac(Gy0, -self.u - 1e-6, self.ref1, self.ref1)
        dae.add_jac(Gy0, -self.u - 1e-6, self.ref2, self.ref2)
        dae.add_jac(Gy0, -self.u - 1e-6, self.vd, self.vd)
        dae.add_jac(Gy0, -self.u - 1e-6, self.vq, self.vq)
        dae.add_jac(Gy0, -self.u - 1e-6, self.p, self.p)
        dae.add_jac(Gy0, -self.u - 1e-6, self.q, self.q)
        dae.add_jac(Gy0, -self.u, self.a, self.p)
        dae.add_jac(Gy0, -self.u, self.v, self.q)


class Power0(object):
    """Empty class for inertia emulation control - place holder"""

    def __init__(self, system, name):
        pass

    def power_init1(self, dae):
        pass

    def power_gcall(self, dae):
        pass

    def power_fcall(self, dae):
        pass

    def power_jac0(self, dae):
        pass

    def power_gycall(self, dae):
        pass

    def power_fxcall(self, dae):
        pass


class Power1(object):
    """
    Inertia emulation control device that modifies Pref based on the
    derivative of frequency deviation
    """

    def __init__(self, system, name):
        self._data.update({
            'Ki': 0,
            'Tf': 1,
        })
        self._descr.update({
            'Ki':
            'droop of the derivative of frequency on active power reference',
            'Tf':
            'dw/dr washout filter time constant'
        })
        self._params.extend({'Ki', 'Tf'})
        self._states.extend(['xdw'])
        self._fnamex.extend(['x_{dw}'])
        self._algebs.extend(['dwdt'])
        self._fnamey.extend(['\\frac{d\\omega}{dt}'])

    def power_init1(self, dae):
        self.iTf = div(self.u, self.Tf)
        dae.x[self.xdw] = self.iTf

    def power_gcall(self, dae):
        dae.g[self.dwdt] = (
            mul(self.iTf, dae.x[self.w]) - dae.x[self.xdw]) - dae.y[self.dwdt]
        dae.g[self.ref1] += -mul(self.Ki, dae.y[self.dwdt])

    def power_fcall(self, dae):
        dae.f[self.xdw] = mul(self.iTf,
                              mul(self.iTf, dae.x[self.w]) - dae.x[self.xdw])

    def power_gycall(self, dae):
        pass

    def power_fxcall(self, dae):
        pass

    def power_jac0(self, dae):
        dae.add_jac(Gx0, self.iTf, self.dwdt, self.w)
        dae.add_jac(Gx0, -1, self.dwdt, self.xdw)

        dae.add_jac(Fx0, self.iTf**2, self.xdw, self.w)
        dae.add_jac(Fx0, -self.iTf, self.xdw, self.xdw)

        dae.add_jac(Gy0, -1, self.dwdt, self.dwdt)

        dae.add_jac(Gy0, -self.Ki, self.ref1, self.dwdt)


class Power2(object):
    """
    Inertia emulation control device that modifies Pref based on the
    derivative of frequency deviation
    """

    def __init__(self, system, name):
        self._data.update({
            'Ki': 0,
            'busfreq': 0,
        })
        self._descr.update({
            'Ki':
            'droop of the derivative of frequency on active power reference',
            'busfreq':
            'BusFreq measurement unit index',
        })
        self._params.extend({'Ki'})

    def power_init1(self, dae):
        self.copy_data_ext('BusFreq', 'dwdt', 'dwdt', self.busfreq)
        self.copy_data_ext('BusFreq', 'w', 'w', self.busfreq)

    def power_gcall(self, dae):
        dae.g[self.ref1] += -mul(self.Ki, dae.y[self.dwdt])

    def power_fcall(self, dae):
        pass

    def power_gycall(self, dae):
        pass

    def power_fxcall(self, dae):
        pass

    def power_jac0(self, dae):
        dae.add_jac(Gy0, -self.Ki, self.ref1, self.dwdt)


class Current1(object):
    """Inner current controllers with two PIs"""

    def __init__(self, system, name):
        self._data.update({
            'Tt': 0.01,
            'Kp1': 0.2,
            'Ki1': 1,
        })
        self._params.extend(['Tt', 'Kp1', 'Ki1'])
        self._states.extend(['ud', 'uq', 'Id', 'Iq', 'Md', 'Mq'])
        self._fnamex.extend(['u_d', 'u_q', 'I_d', 'I_q', 'M_d', 'M_q'])
        self._descr.update({
            'Ki1':
            'Innercurrent controller integrator gain',
            'Kp1':
            'Inner current controller proportional gain',
            'Tt':
            'ac voltage measurement delay time constant',
        })
        self._service.extend(['iTt', 'iLsh'])

    def current_servcall(self, dae):
        self.iTt = div(1, self.Tt)
        self.iLsh = div(1, self.xsh)

    def current_init1(self, dae):
        self.current_servcall(dae)
        dae.x[self.Id] = mul(self.pref0, div(1, dae.y[self.vd]))
        dae.x[self.Iq] = mul(self.qref0, div(1, dae.y[self.vd]))
        dae.x[self.ud] = dae.y[self.vd] + mul(dae.x[self.Id], self.rsh) + mul(
            dae.x[self.Iq], self.xsh)
        dae.x[self.uq] = dae.y[self.vq] + mul(dae.x[self.Iq], self.rsh) - mul(
            dae.x[self.Id], self.xsh)
        dae.x[self.Md] = mul(dae.x[self.Id], self.rsh)
        dae.x[self.Mq] = mul(dae.x[self.Iq], self.rsh)

    def current_fcall(self, dae):
        dae.f[self.Id] = -dae.x[self.Iq] + mul(
            self.iLsh, dae.x[self.ud] - dae.y[self.vd]) - mul(
                dae.x[self.Id], self.iLsh, self.rsh)
        dae.f[self.Iq] = dae.x[self.Id] + mul(
            self.iLsh, dae.x[self.uq] - dae.y[self.vq]) - mul(
                dae.x[self.Iq], self.iLsh, self.rsh)
        dae.f[self.ud] = mul(
            self.iTt, dae.x[self.Md] + dae.y[self.vd] - dae.x[self.ud] + mul(
                dae.y[self.Iqref], self.xsh) + mul(
                    self.Kp1, dae.y[self.Idref] - dae.x[self.Id]))
        dae.f[self.uq] = mul(
            self.iTt, dae.x[self.Mq] + dae.y[self.vq] - dae.x[self.uq] + mul(
                self.Kp1, dae.y[self.Iqref] - dae.x[self.Iq]) - mul(
                    dae.y[self.Idref], self.xsh))
        dae.f[self.Md] = mul(self.Ki1, dae.y[self.Idref] - dae.x[self.Id])
        dae.f[self.Mq] = mul(self.Ki1, dae.y[self.Iqref] - dae.x[self.Iq])

    def current_jac0(self, dae):
        dae.add_jac(Fx0, -1, self.Id, self.Iq)
        dae.add_jac(Fx0, self.iLsh, self.Id, self.ud)
        dae.add_jac(Fx0, -mul(self.iLsh, self.rsh), self.Id, self.Id)
        dae.add_jac(Fx0, -mul(self.iLsh, self.rsh), self.Iq, self.Iq)
        dae.add_jac(Fx0, self.iLsh, self.Iq, self.uq)
        dae.add_jac(Fx0, 1, self.Iq, self.Id)
        dae.add_jac(Fx0, -self.iTt, self.ud, self.ud)
        dae.add_jac(Fx0, -mul(self.Kp1, self.iTt), self.ud, self.Id)
        dae.add_jac(Fx0, self.iTt, self.ud, self.Md)
        dae.add_jac(Fx0, self.iTt, self.uq, self.Mq)
        dae.add_jac(Fx0, -self.iTt, self.uq, self.uq)
        dae.add_jac(Fx0, -mul(self.Kp1, self.iTt), self.uq, self.Iq)
        dae.add_jac(Fx0, -self.Ki1, self.Md, self.Id)
        dae.add_jac(Fx0, -self.Ki1, self.Mq, self.Iq)

        dae.add_jac(Fy0, -self.iLsh, self.Id, self.vd)
        dae.add_jac(Fy0, -self.iLsh, self.Iq, self.vq)
        dae.add_jac(Fy0, mul(self.Kp1, self.iTt), self.ud, self.Idref)
        dae.add_jac(Fy0, mul(self.iTt, self.xsh), self.ud, self.Iqref)
        dae.add_jac(Fy0, self.iTt, self.ud, self.vd)
        dae.add_jac(Fy0, -mul(self.iTt, self.xsh), self.uq, self.Idref)
        dae.add_jac(Fy0, mul(self.Kp1, self.iTt), self.uq, self.Iqref)
        dae.add_jac(Fy0, self.iTt, self.uq, self.vq)
        dae.add_jac(Fy0, self.Ki1, self.Md, self.Idref)
        dae.add_jac(Fy0, self.Ki1, self.Mq, self.Iqref)

    def current_gcall(self, dae):
        pass

    def current_gycall(self, dae):
        pass

    def current_fxcall(self, dae):
        pass


class VSC1_Outer1(object):
    """Outer power control loop for VSC1"""

    def __init__(self, system, name):
        self._data.update({
            'Ki2': 1,
            'Ki3': 1,
            'Kp2': 0.2,
            'Kp3': 0.2,
            'Tdc': 0.01,
        })
        self._descr.update({
            'Ki2':
            'P or Vdc voltage controller integrator gain',
            'Ki3':
            'Q or V controller integrator gain',
            'Kp2':
            'P or Vdc controller proportional gain',
            'Kp3':
            'Q or V controller proportional gain',
            'Tdc':
            'dc voltage time constant',
        })
        self._params.extend(['Kp2', 'Ki2', 'Kp3', 'Ki3', 'Tdc'])

        self._algebs.extend(['Idref', 'Iqref', 'Idcy'])
        self._fnamey.extend(['I_d^{ref}', 'I_q^{ref}', 'I_{dcy}'])
        self._states.extend(['Nd', 'Nq', 'Idcx'])
        self._fnamex.extend(['N_d', 'N_q', 'I_{dcx}'])
        self._service.extend(['iTdc'])

    def outer_servcall(self, dae):
        self.iTdc = div(1, self.Tdc)

    def outer_init1(self, dae):
        self.outer_servcall(dae)
        dae.y[self.Idref] = mul(self.pref0, div(1, dae.y[self.vd]))
        dae.y[self.Iqref] = mul(self.qref0, div(1, dae.y[self.vd]))
        dae.y[self.Idcy] = mul(
            div(1, dae.y[self.v1] - dae.y[self.v2]),
            self.PQ + self.PV, -mul(dae.x[self.Id], dae.x[self.ud]) - mul(
                dae.x[self.Iq], dae.x[self.uq]))
        dae.x[self.Idcx] = mul(
            div(1, dae.y[self.v1] - dae.y[self.v2]),
            self.vQ + self.vV, -mul(dae.x[self.Id], dae.x[self.ud]) - mul(
                dae.x[self.Iq], dae.x[self.uq]))
        dae.x[self.Nd] = mul(dae.x[self.Idcx], self.vQ + self.vV)
        dae.x[self.Nq] = mul(dae.x[self.Iq], self.PV + self.vV)

    def outer_gcall(self, dae):
        dae.g[self.Idref] = -dae.y[self.Idref] + \
            mul(self.PQ + self.PV,
                dae.x[self.Nd] + mul(self.Kp2,
                                     dae.y[self.ref1] - dae.y[
                                         self.p]) + mul(
                    dae.y[self.ref1], div(1, dae.y[self.vd]))) + \
            mul(div(1, dae.x[self.ud]), self.vQ + self.vV,
                -mul(dae.x[self.Idcx],
                     dae.y[self.v1] - dae.y[self.v2]) - mul(
                    dae.x[self.Iq], dae.x[self.uq]))

        dae.g[self.Iqref] = dae.x[self.Nq] - dae.y[self.Iqref] + mul(
            self.PQ + self.vQ,
            mul(self.Kp3, dae.y[self.ref2] - dae.y[self.q]) +
            mul(dae.y[self.ref2], div(1, dae.y[self.vd]))) + mul(
                self.Kp3, self.PV + self.vV, dae.y[self.ref2] - dae.y[self.vd])
        dae.g[self.Idcy] = -dae.y[self.Idcy] + mul(
            div(1, dae.y[self.v1] - dae.y[self.v2]),
            self.PQ + self.PV, -mul(dae.x[self.Id], dae.x[self.ud]) - mul(
                dae.x[self.Iq], dae.x[self.uq]))
        dae.g += spmatrix(
            mul(dae.y[self.Idcy], -self.PQ - self.PV) - mul(
                dae.x[self.Idcx], self.vQ + self.vV), self.v1, [0] * self.n,
            (dae.m, 1), 'd')
        dae.g += spmatrix(
            mul(dae.x[self.Idcx], self.vQ + self.vV) + mul(
                dae.y[self.Idcy], self.PQ + self.PV), self.v2, [0] * self.n,
            (dae.m, 1), 'd')

    def outer_fcall(self, dae):
        dae.f[self.Nd] = mul(self.Ki2, self.PQ + self.PV,
                             dae.y[self.ref1] - dae.y[self.p]) + mul(
                                 self.Ki2, dae.y[self.ref1] - dae.y[self.v1],
                                 self.vQ + self.vV)
        dae.f[self.Nq] = mul(self.Ki3, self.PQ + self.vQ,
                             dae.y[self.ref2] - dae.y[self.q]) + mul(
                                 self.Ki3, self.PV + self.vV,
                                 dae.y[self.ref2] - dae.y[self.vd])
        dae.f[self.Idcx] = mul(
            self.iTdc, self.vQ + self.vV,
            -dae.x[self.Idcx] + dae.x[self.Nd] + mul(
                self.Kp2, dae.y[self.ref1] - dae.y[self.v1]))

    def outer_gycall(self, dae):
        dae.add_jac(Gy,
                    mul(self.Kp2 + div(1, dae.y[self.vd]), self.PQ + self.PV),
                    self.Idref, self.ref1)
        dae.add_jac(Gy,
                    mul(dae.x[self.Idcx], div(1, dae.x[self.ud]),
                        self.vQ + self.vV), self.Idref, self.v2)  # beware
        dae.add_jac(
            Gy,
            -mul(dae.x[self.Idcx], div(1, dae.x[self.ud]), self.vQ + self.vV),
            self.Idref, self.v1)
        dae.add_jac(
            Gy, -mul(dae.y[self.ref1],
                     (dae.y[self.vd])**-2, self.PQ + self.PV), self.Idref,
            self.vd)

        dae.add_jac(
            Gy,
            mul(self.Kp3, self.PV + self.vV) + mul(
                self.Kp3 + div(1, dae.y[self.vd]), self.PQ + self.vQ),
            self.Iqref, self.ref2)
        dae.add_jac(
            Gy, -mul(self.Kp3, self.PV + self.vV) - mul(
                dae.y[self.ref2], (dae.y[self.vd])**-2, self.PQ + self.vQ),
            self.Iqref, self.vd)
        dae.add_jac(
            Gy,
            mul((dae.y[self.v1] - dae.y[self.v2])**-2, self.PQ + self.PV,
                -mul(dae.x[self.Id], dae.x[self.ud]) - mul(
                    dae.x[self.Iq], dae.x[self.uq])), self.Idcy, self.v2)
        dae.add_jac(
            Gy, -mul((dae.y[self.v1] - dae.y[self.v2])**-2, self.PQ + self.PV,
                     -mul(dae.x[self.Id], dae.x[self.ud]) - mul(
                         dae.x[self.Iq], dae.x[self.uq])), self.Idcy, self.v1)

    def outer_fxcall(self, dae):
        dae.add_jac(
            Gx,
            mul(
                div(1, dae.x[self.ud]), dae.y[self.v2] - dae.y[self.v1],
                self.vQ + self.vV), self.Idref, self.Idcx)
        dae.add_jac(
            Gx,
            -mul((dae.x[self.ud])**-2, self.vQ + self.vV,
                 -mul(dae.x[self.Idcx], dae.y[self.v1] - dae.y[self.v2]) - mul(
                     dae.x[self.Iq], dae.x[self.uq])), self.Idref, self.ud)
        dae.add_jac(
            Gx,
            -mul(dae.x[self.Iq], div(1, dae.x[self.ud]), self.vQ + self.vV),
            self.Idref, self.uq)
        dae.add_jac(
            Gx,
            -mul(dae.x[self.uq], div(1, dae.x[self.ud]), self.vQ + self.vV),
            self.Idref, self.Iq)

        dae.add_jac(
            Gx, -mul(dae.x[self.Id], div(1, dae.y[self.v1] - dae.y[self.v2]),
                     self.PQ + self.PV), self.Idcy, self.ud)
        dae.add_jac(
            Gx, -mul(dae.x[self.Iq], div(1, dae.y[self.v1] - dae.y[self.v2]),
                     self.PQ + self.PV), self.Idcy, self.uq)
        dae.add_jac(
            Gx, -mul(dae.x[self.ud], div(1, dae.y[self.v1] - dae.y[self.v2]),
                     self.PQ + self.PV), self.Idcy, self.Id)
        dae.add_jac(
            Gx, -mul(dae.x[self.uq], div(1, dae.y[self.v1] - dae.y[self.v2]),
                     self.PQ + self.PV), self.Idcy, self.Iq)

        dae.add_jac(Fx, -mul(self.iTdc, self.vQ + self.vV), self.Idcx,
                    self.Idcx)
        dae.add_jac(Fx, mul(self.iTdc, self.vQ + self.vV), self.Idcx, self.Nd)
        dae.add_jac(Fy, mul(self.Kp2, self.iTdc, self.vQ + self.vV), self.Idcx,
                    self.ref1)
        dae.add_jac(Fy, -mul(self.Kp2, self.iTdc, self.vQ + self.vV),
                    self.Idcx, self.v1)

    def outer_jac0(self, dae):
        dae.add_jac(Gy0, -1, self.Idref, self.Idref)
        dae.add_jac(Gy0, -mul(self.Kp2, self.PQ + self.PV), self.Idref, self.p)
        dae.add_jac(Gy0, -mul(self.Kp3, self.PQ + self.vQ), self.Iqref, self.q)
        dae.add_jac(Gy0, -1, self.Iqref, self.Iqref)
        dae.add_jac(Gy0, -1, self.Idcy, self.Idcy)
        dae.add_jac(Gy0, -self.PQ - self.PV, self.v1, self.Idcy)
        dae.add_jac(Gy0, self.PQ + self.PV, self.v2, self.Idcy)
        dae.add_jac(Gx0, self.PQ + self.PV, self.Idref, self.Nd)
        dae.add_jac(Gx0, 1, self.Iqref, self.Nq)
        dae.add_jac(Gx0, -self.vQ - self.vV, self.v1, self.Idcx)
        dae.add_jac(Gx0, self.vQ + self.vV, self.v2, self.Idcx)

        dae.add_jac(
            Fy0,
            mul(self.Ki2, self.PQ + self.PV) + mul(
                self.Ki2, self.vQ + self.vV), self.Nd, self.ref1)
        dae.add_jac(Fy0, -mul(self.Ki2, self.vQ + self.vV), self.Nd, self.v1)
        dae.add_jac(Fy0, -mul(self.Ki2, self.PQ + self.PV), self.Nd, self.p)
        dae.add_jac(
            Fy0,
            mul(self.Ki3, self.PQ + self.vQ) + mul(
                self.Ki3, self.PV + self.vV), self.Nq, self.ref2)
        dae.add_jac(Fy0, -mul(self.Ki3, self.PV + self.vV), self.Nq, self.vd)
        dae.add_jac(Fy0, -mul(self.Ki3, self.PQ + self.vQ), self.Nq, self.q)
        dae.add_jac(Gy0, 1e-6, self.Idref, self.Idref)
        dae.add_jac(Gy0, 1e-6, self.Iqref, self.Iqref)
        dae.add_jac(Gy0, 1e-6, self.Idcy, self.Idcy)


class PLL1(object):
    """Ideal tracking PLL"""

    def __init__(self, system, name):
        self._algebs.extend(['adq'])
        self._params.extend(['adq'])
        self._fnamey.extend(['\\theta{dq}'])

    def pll_init1(self, dae):
        dae.y[self.adq] = dae.y[self.a]

    def pll_gcall(self, dae):
        dae.g[self.adq] = dae.y[self.adq] - dae.y[self.a]

    def pll_jac0(self, dae):
        dae.add_jac(Gy0, -1, self.adq, self.a)
        dae.add_jac(Gy0, 1 + 1e-6, self.adq, self.adq)

    def pll_fcall(self, dae):
        pass

    def pll_fxcall(self, dae):
        pass

    def pll_gycall(self, dae):
        pass


class VSC1(VSC1_Common, VSC1_Outer1, Current1, PLL1, Power0):
    def __init__(self, system, name):
        VSC1_Common.__init__(self, system, name)
        VSC1_Outer1.__init__(self, system, name)
        Current1.__init__(self, system, name)
        PLL1.__init__(self, system, name)
        Power0.__init__(self, system, name)
        self._init()


class VSC1_IE(VSC1_Common, VSC1_Outer1, Current1, PLL1, Power1):
    """VSC1 with Inertia Emulation with frequency sensor at local bus"""

    def __init__(self, system, name):
        VSC1_Common.__init__(self, system, name)
        VSC1_Outer1.__init__(self, system, name)
        Current1.__init__(self, system, name)
        PLL1.__init__(self, system, name)
        Power1.__init__(self, system, name)
        self._init()


class VSC1_IE2(VSC1_Common, VSC1_Outer1, Current1, PLL1, Power2):
    """VSC1 with Inertia Emulation with frequency sensor at remote bus"""

    def __init__(self, system, name):
        VSC1_Common.__init__(self, system, name)
        VSC1_Outer1.__init__(self, system, name)
        Current1.__init__(self, system, name)
        PLL1.__init__(self, system, name)
        Power2.__init__(self, system, name)
        self._init()


class VSC2_Voltage1(object):
    """
    Outer voltage controller for voltage-source controlled VSC using two PIs
    """

    def __init__(self, system, name):
        self._data.update({
            'Ki2': 1,
            'Ki3': 1,
            'Kp2': 0.2,
            'Kp3': 0.2,
            'KQ': 0.1
        })
        self._descr.update({
            'Ki2': 'vd -> v controller integrator gain',
            'Ki3': 'vq -> 0 controller integrator gain',
            'Kp2': 'vd -> v controller proportional gain',
            'Kp3': 'vq -> 0 controller proportional gain',
            'KQ': 'reactive power droop on voltage',
        })
        self._params.extend(['Ki2', 'Ki3', 'Kp2', 'Kp3', 'KQ'])

        self._algebs.extend(['ref2', 'Idref', 'Iqref'])
        self._fnamey.extend(['ref_2', 'I_d^{ref}', 'I_q^{ref}'])

        self._states.extend(['Nd', 'Nq'])
        self._fnamex.extend(['N_d', 'N_q'])

    def voltage_init1(self, dae):
        dae.y[self.Idref] = mul(self.pref0, div(1, dae.y[self.vd]))
        dae.y[self.Iqref] = mul(self.qref0, div(1, dae.y[self.vd]))
        dae.x[self.Nd] = dae.y[self.Idref]
        dae.x[self.Nq] = dae.y[self.Iqref]
        dae.y[self.ref2] = dae.y[self.v]

    def voltage_gcall(self, dae):
        dae.g[self.ref2] = self.vref0 - dae.y[self.ref2] + mul(
            self.KQ, self.qref0 - dae.y[self.q])
        dae.g[self.Idref] = dae.x[self.Nd] - dae.y[self.Idref] + mul(
            self.Kp2, dae.y[self.ref2] - dae.y[self.vd])
        dae.g[self.Iqref] = dae.x[self.Nq] - dae.y[self.Iqref] - mul(
            self.Kp3, dae.y[self.vq])

    def voltage_fcall(self, dae):
        dae.f[self.Nd] = mul(self.Ki2, self.u,
                             dae.y[self.ref2] - dae.y[self.vd])
        dae.f[self.Nq] = -mul(self.Ki3, self.u, dae.y[self.vq])

    def voltage_gycall(self, dae):
        pass

    def voltage_fxcall(self, dae):
        pass

    def voltage_jac0(self, dae):
        dae.add_jac(Gy0, -self.u - 1e-6, self.ref2, self.ref2)
        dae.add_jac(Gy0, -self.KQ, self.ref2, self.q)
        dae.add_jac(Gy0, self.Kp2, self.Idref, self.ref2)
        dae.add_jac(Gy0, -self.Kp2, self.Idref, self.vd)
        dae.add_jac(Gy0, -self.u - 1e-6, self.Idref, self.Idref)
        dae.add_jac(Gy0, -self.u - 1e-6, self.Iqref, self.Iqref)
        dae.add_jac(Gy0, -self.Kp3, self.Iqref, self.vq)

        dae.add_jac(Gx0, 1, self.Idref, self.Nd)
        dae.add_jac(Gx0, 1, self.Iqref, self.Nq)

        dae.add_jac(Fy0, mul(self.Ki2, self.u), self.Nd, self.ref2)
        dae.add_jac(Fy0, -mul(self.Ki2, self.u), self.Nd, self.vd)
        dae.add_jac(Fy0, -mul(self.Ki3, self.u), self.Nq, self.vq)


class VSC2_Speed1(object):
    """Active power droop speed control"""

    def __init__(self, system, name):
        self._data.update({
            'D': 0.5,
        })
        self._descr.update({
            'D': 'Active power droop for speed reference',
        })
        self._params.extend(['D'])
        self._algebs.extend(['ref1'])
        self._fnamey.extend(['ref_1'])

        self._states.extend(['adq'])
        self._fnamex.extend(['\\theta_{dq}'])
        self._powers.extend(['D'])

    def speed_init1(self, dae):
        self.iD = div(self.u, self.D)
        dae.x[self.adq] = mul(dae.y[self.a], self.u)
        dae.y[self.ref1] = self.wref0

    def speed_gcall(self, dae):
        dae.g[self.ref1] = self.wref0 - dae.y[self.ref1] + mul(
            self.iD, self.pref0 - dae.y[self.p])

    def speed_fcall(self, dae):
        dae.f[self.adq] = mul(self.u, dae.y[self.ref1] - self.wref0)

    def speed_fxcall(self, dae):
        pass

    def speed_gycall(self, dae):
        pass

    def speed_jac0(self, dae):
        dae.add_jac(Gy0, -self.u - 1e-6, self.ref1, self.ref1)
        dae.add_jac(Gy0, -self.iD, self.ref1, self.p)
        dae.add_jac(Fy0, self.u, self.adq, self.ref1)


class VSC2_Speed2(object):
    """Inertia emulation speed control"""

    def __init__(self, system, name):
        self._data.update({
            'D': 0.5,
            'M': 3,
        })
        self._descr.update({
            'D': 'Emulated damping',
            'M': 'Emulated start-up time constant',
        })
        self._params.extend(['M', 'D'])
        self._algebs.extend(['ref1'])
        self._fnamey.extend(['ref_1'])

        self._states.extend(['adq', 'xw'])
        self._fnamex.extend(['\\theta_{dq}', 'x_\\omega'])

        self._service.extend(['iM'])
        self._powers.extend(['M', 'D'])

    def speed_servcall(self, dae):
        self.iM = mul(self.u, div(1, self.M))
        # self.iD = div(self.u, self.D)

    def speed_init1(self, dae):
        self.speed_servcall(dae)
        dae.x[self.adq] = mul(dae.y[self.a], self.u)
        dae.y[self.ref1] = self.wref0

    def speed_gcall(self, dae):
        dae.g[self.ref1] = self.wref0 + dae.x[self.xw] - dae.y[self.ref1]

    def speed_fcall(self, dae):
        dae.f[self.adq] = mul(self.u, dae.y[self.ref1] - self.wref0)
        dae.f[self.xw] = mul(
            self.iM, self.u,
            self.pref0 - dae.y[self.p] - mul(self.D, dae.x[self.xw]))

    def speed_fxcall(self, dae):
        pass

    def speed_gycall(self, dae):
        pass

    def speed_jac0(self, dae):
        dae.add_jac(Gy0, -self.u - 1e-6, self.ref1, self.ref1)
        dae.add_jac(Gx0, self.u, self.ref1, self.xw)
        dae.add_jac(Fy0, self.u, self.adq, self.ref1)
        dae.add_jac(Fx0, -mul(self.D, self.iM, self.u), self.xw, self.xw)
        dae.add_jac(Fy0, -mul(self.iM, self.u), self.xw, self.p)


class VSC2_Common(DCBase):
    """Common equations for voltage-source controlled VSC"""

    def __init__(self, system, name):
        super(VSC2_Common, self).__init__(system, name)
        self._group = 'AC/DC'
        self._data.update({
            'vsc': None,
        })
        self._descr.update({'vsc': 'static vsc idx'})
        self._params.extend(['vsc', 'Sn'])
        self._mandatory.extend(['vsc'])

        self._algebs.extend(['p', 'q', 'vd', 'vq'])
        self._fnamey.extend(['P', 'Q', 'v_d', 'v_q'])
        self._service.extend(['qref0', 'vref0', 'wref0', 'pref0'])

        self.calls.update({
            'gcall': True,
            'jac0': True,
            'gycall': True,
            'init1': True,
            'fxcall': True,
            'fcall': True
        })
        self._mandatory.remove('node1')
        self._mandatory.remove('node2')
        self._mandatory.remove('Vdcn')
        self._ac = {}
        self._dc = {}

    def servcall(self, dae):
        self.copy_data_ext('VSC', 'u', 'uvsc', self.vsc)
        self.copy_data_ext('VSC', 'rsh', 'rsh', self.vsc)
        self.copy_data_ext('VSC', 'xsh', 'xsh', self.vsc)
        self.copy_data_ext('VSC', 'PQ', 'PQ', self.vsc)
        self.copy_data_ext('VSC', 'PV', 'PV', self.vsc)
        self.copy_data_ext('VSC', 'vQ', 'vQ', self.vsc)
        self.copy_data_ext('VSC', 'vV', 'vV', self.vsc)
        self.copy_data_ext('VSC', 'a', 'a', self.vsc)
        self.copy_data_ext('VSC', 'v', 'v', self.vsc)
        self.copy_data_ext('VSC', 'v1', 'v1', self.vsc)
        self.copy_data_ext('VSC', 'v2', 'v2', self.vsc)
        self.copy_data_ext('VSC', 'psh', 'psh', self.vsc)
        self.copy_data_ext('VSC', 'qsh', 'qsh', self.vsc)
        self.u = mul(self.u, self.uvsc)
        self.vref0 = mul(self.u, dae.y[self.v])
        self.wref0 = self.u
        self.pref0 = mul(dae.y[self.psh], self.u)
        self.qref0 = mul(dae.y[self.qsh], self.u)

    def init1(self, dae):
        self.servcall(dae)
        dae.y[self.vd] = mul(self.u, dae.y[self.v])
        dae.y[self.vq] = 0
        dae.x[self.Id] = mul(self.pref0, div(1, dae.y[self.vd]))
        dae.x[self.Iq] = mul(self.qref0, div(1, dae.y[self.vd]))
        dae.y[self.p] = mul(dae.x[self.Id], dae.y[self.vd]) + mul(
            dae.x[self.Iq], dae.y[self.vq])
        dae.y[self.q] = mul(dae.x[self.Iq], dae.y[self.vd]) - mul(
            dae.x[self.Id], dae.y[self.vq])

        self.speed_init1(dae)
        self.voltage_init1(dae)
        self.current_init1(dae)

        for idx in self.vsc:
            self.system.VSC.disable(idx)

    def gcall(self, dae):
        self.speed_gcall(dae)
        self.voltage_gcall(dae)
        self.current_gcall(dae)
        dae.g[self.p] = -dae.y[self.p] + mul(
            dae.x[self.Id], dae.y[self.vd]) + mul(dae.x[self.Iq],
                                                  dae.y[self.vq])
        dae.g[self.q] = -dae.y[self.q] + mul(
            dae.x[self.Iq], dae.y[self.vd]) - mul(dae.x[self.Id],
                                                  dae.y[self.vq])
        dae.g[self.vd] = -dae.y[self.vd] + mul(
            self.u, dae.y[self.v], cos(dae.y[self.a] - dae.x[self.adq]))
        dae.g[self.vq] = -dae.y[self.vq] - mul(
            self.u, dae.y[self.v], sin(dae.y[self.a] - dae.x[self.adq]))
        dae.g += spmatrix(-mul(dae.y[self.p], self.u), self.a, [0] * self.n,
                          (dae.m, 1), 'd')
        dae.g += spmatrix(-mul(dae.y[self.q], self.u), self.v, [0] * self.n,
                          (dae.m, 1), 'd')
        dae.g += spmatrix(
            mul(
                self.u, div(1, dae.y[self.v1] - dae.y[self.v2]),
                mul(dae.x[self.Id], dae.x[self.ud]) + mul(
                    dae.x[self.Iq], dae.x[self.uq])), self.v1, [0] * self.n,
            (dae.m, 1), 'd')
        dae.g += spmatrix(
            -mul(
                self.u, div(1, dae.y[self.v1] - dae.y[self.v2]),
                mul(dae.x[self.Id], dae.x[self.ud]) + mul(
                    dae.x[self.Iq], dae.x[self.uq])), self.v2, [0] * self.n,
            (dae.m, 1), 'd')

    def fcall(self, dae):
        self.speed_fcall(dae)
        self.voltage_fcall(dae)
        self.current_fcall(dae)

    def gycall(self, dae):
        self.speed_gycall(dae)
        self.voltage_gycall(dae)
        self.current_gycall(dae)

        dae.add_jac(Gy, dae.x[self.Iq], self.p, self.vq)
        dae.add_jac(Gy, dae.x[self.Id], self.p, self.vd)
        dae.add_jac(Gy, -dae.x[self.Id], self.q, self.vq)
        dae.add_jac(Gy, dae.x[self.Iq], self.q, self.vd)
        dae.add_jac(
            Gy,
            -mul(self.u, dae.y[self.v], sin(dae.y[self.a] - dae.x[self.adq])),
            self.vd, self.a)
        dae.add_jac(Gy, mul(self.u, cos(dae.y[self.a] - dae.x[self.adq])),
                    self.vd, self.v)
        dae.add_jac(Gy, -mul(self.u, sin(dae.y[self.a] - dae.x[self.adq])),
                    self.vq, self.v)
        dae.add_jac(
            Gy,
            -mul(self.u, dae.y[self.v], cos(dae.y[self.a] - dae.x[self.adq])),
            self.vq, self.a)
        dae.add_jac(
            Gy,
            mul(
                self.u, (dae.y[self.v1] - dae.y[self.v2])**-2,
                mul(dae.x[self.Id], dae.x[self.ud]) + mul(
                    dae.x[self.Iq], dae.x[self.uq])), self.v1, self.v2)
        dae.add_jac(
            Gy, -mul(
                self.u, (dae.y[self.v1] - dae.y[self.v2])**-2,
                mul(dae.x[self.Id], dae.x[self.ud]) + mul(
                    dae.x[self.Iq], dae.x[self.uq])), self.v1, self.v1)
        dae.add_jac(
            Gy, -mul(
                self.u, (dae.y[self.v1] - dae.y[self.v2])**-2,
                mul(dae.x[self.Id], dae.x[self.ud]) + mul(
                    dae.x[self.Iq], dae.x[self.uq])), self.v2, self.v2)
        dae.add_jac(
            Gy,
            mul(
                self.u, (dae.y[self.v1] - dae.y[self.v2])**-2,
                mul(dae.x[self.Id], dae.x[self.ud]) + mul(
                    dae.x[self.Iq], dae.x[self.uq])), self.v2, self.v1)

    def fxcall(self, dae):
        self.speed_fxcall(dae)
        self.voltage_fxcall(dae)
        self.current_fxcall(dae)

        dae.add_jac(Gx, dae.y[self.vq], self.p, self.Iq)
        dae.add_jac(Gx, dae.y[self.vd], self.p, self.Id)
        dae.add_jac(Gx, dae.y[self.vd], self.q, self.Iq)
        dae.add_jac(Gx, -dae.y[self.vq], self.q, self.Id)
        dae.add_jac(
            Gx, mul(self.u, dae.y[self.v],
                    sin(dae.y[self.a] - dae.x[self.adq])), self.vd, self.adq)
        dae.add_jac(
            Gx, mul(self.u, dae.y[self.v],
                    cos(dae.y[self.a] - dae.x[self.adq])), self.vq, self.adq)
        dae.add_jac(
            Gx,
            mul(self.u, dae.x[self.uq],
                div(1, dae.y[self.v1] - dae.y[self.v2])), self.v1, self.Iq)
        dae.add_jac(
            Gx,
            mul(self.u, dae.x[self.ud],
                div(1, dae.y[self.v1] - dae.y[self.v2])), self.v1, self.Id)
        dae.add_jac(
            Gx,
            mul(dae.x[self.Id], self.u,
                div(1, dae.y[self.v1] - dae.y[self.v2])), self.v1, self.ud)
        dae.add_jac(
            Gx,
            mul(dae.x[self.Iq], self.u,
                div(1, dae.y[self.v1] - dae.y[self.v2])), self.v1, self.uq)
        dae.add_jac(
            Gx, -mul(self.u, dae.x[self.uq],
                     div(1, dae.y[self.v1] - dae.y[self.v2])), self.v2,
            self.Iq)
        dae.add_jac(
            Gx, -mul(self.u, dae.x[self.ud],
                     div(1, dae.y[self.v1] - dae.y[self.v2])), self.v2,
            self.Id)
        dae.add_jac(
            Gx, -mul(dae.x[self.Id], self.u,
                     div(1, dae.y[self.v1] - dae.y[self.v2])), self.v2,
            self.ud)
        dae.add_jac(
            Gx, -mul(dae.x[self.Iq], self.u,
                     div(1, dae.y[self.v1] - dae.y[self.v2])), self.v2,
            self.uq)

    def jac0(self, dae):
        self.speed_jac0(dae)
        self.voltage_jac0(dae)
        self.current_jac0(dae)

        dae.add_jac(Gy0, -1, self.p, self.p)
        dae.add_jac(Gy0, -1, self.q, self.q)
        dae.add_jac(Gy0, -1, self.vd, self.vd)
        dae.add_jac(Gy0, -1, self.vq, self.vq)
        dae.add_jac(Gy0, -self.u, self.a, self.p)
        dae.add_jac(Gy0, -self.u, self.v, self.q)


class VSC2A(VSC2_Common, Current1, VSC2_Speed1, VSC2_Voltage1):
    """Voltage-source controlled VSC with active power droop"""

    def __init__(self, system, name):
        VSC2_Common.__init__(self, system, name)
        Current1.__init__(self, system, name)
        VSC2_Speed1.__init__(self, system, name)
        VSC2_Voltage1.__init__(self, system, name)
        self._name = 'VSC2A'
        self._init()


class VSC2B(VSC2_Common, Current1, VSC2_Speed2, VSC2_Voltage1):
    """Voltage-source controlled VSC with inertia emulation"""

    def __init__(self, system, name):
        VSC2_Common.__init__(self, system, name)
        Current1.__init__(self, system, name)
        VSC2_Speed2.__init__(self, system, name)
        VSC2_Voltage1.__init__(self, system, name)
        self._name = 'VSC2B'
        self._init()



"""PSS/E file parser"""

import re

from ..consts import deg2rad
from ..utils.math import to_number
import logging

logger = logging.getLogger(__name__)


def testlines(fid):
    """Check the raw file for frequency base"""
    first = fid.readline()
    first = first.strip().split('/')
    first = first[0].split(',')
    if float(first[5]) == 50.0 or float(first[5]) == 60.0:
        return True
    else:
        return False


def get_nol(b, mdata):
    """
    Return the number of lines based on data
    """
    nol = [1, 1, 1, 1, 1, 4, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0]

    if b == 5:  # for transformer
        if mdata[0][2] == 0:  # two-winding transformer
            return 4
        else:  # three-widning transf
            return 5

    return nol[b]


def read(file, system):
    """read PSS/E RAW file v32 format"""

    blocks = [
        'bus', 'load', 'fshunt', 'gen', 'branch', 'transf', 'area',
        'twotermdc', 'vscdc', 'impedcorr', 'mtdc', 'msline', 'zone',
        'interarea', 'owner', 'facts', 'swshunt', 'gne', 'Q'
    ]
    nol = [1, 1, 1, 1, 1, 4, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0]
    rawd = re.compile(r'rawd\d\d')

    retval = True
    version = 0
    b = 0  # current block index
    raw = {}
    for item in blocks:
        raw[item] = []

    data = []
    mdata = []  # multi-line data
    mline = 0  # line counter for multi-line models

    # parse file into raw with to_number conversions
    fid = open(file, 'r')
    for num, line in enumerate(fid.readlines()):
        line = line.strip()
        if num == 0:  # get basemva and frequency
            data = line.split('/')[0]
            data = data.split(',')

            mva = float(data[1])
            system.config.mva = mva
            system.config.freq = float(data[5])

            # get raw file version
            version = int(data[2])
            if not version:
                version = int(rawd.search(line).group(0).strip('rawd'))
            logger.debug('PSSE raw version {} detected'.format(version))

            if version < 32 or version > 33:
                logger.warning(
                    'RAW file version is not 32 or 33. Error may occur.')
            continue

        elif num == 1:  # store the case info line
            logger.info(line)
            continue
        elif num == 2:
            continue
        elif num >= 3:
            if line[0:2] == '0 ' or line[0:3] == ' 0 ':  # end of block
                b += 1
                continue
            elif line[0] == 'Q':  # end of file
                break
            data = line.split(',')

        data = [to_number(item) for item in data]
        mdata.append(data)
        mline += 1
        nol = get_nol(b, mdata)

        if mline >= nol:
            if nol == 1:
                mdata = mdata[0]
            raw[blocks[b]].append(mdata)
            mdata = []
            mline = 0
    fid.close()

    # add device elements to system
    sw = {}  # idx:a0
    max_bus = []
    for data in raw['bus']:
        """version 32:
          0,   1,      2,     3,    4,   5,  6,   7,  8
          ID, NAME, BasekV, Type, Area Zone Owner Vm, Va
        """
        idx = data[0]
        max_bus.append(idx)
        ty = data[3]
        a0 = data[8] * deg2rad

        if ty == 3:
            sw[idx] = a0
        param = {
            'idx': idx,
            'name': data[1],
            'Vn': data[2],
            'voltage': data[7],
            'angle': a0,
            'area': data[4],
            'zone': data[5],
            'owner': data[6],
        }
        system.Bus.elem_add(**param)
    max_bus = max(max_bus)
    for data in raw['load']:
        """version 32:
          0,  1,      2,    3,    4,    5,    6,      7,   8,  9, 10,   11
        Bus, Id, Status, Area, Zone, PL(MW), QL (MW), IP, IQ, YP, YQ, OWNER
        """
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        voltage = system.Bus.get_field('voltage', bus)
        param = {
            'bus': bus,
            'Vn': vn,
            'Sn': mva,
            'p': (data[5] + data[7] * voltage + data[9] * voltage**2) / mva,
            'q': (data[6] + data[8] * voltage - data[10] * voltage**2) / mva,
            'owner': data[11],
        }
        system.PQ.elem_add(**param)

    for data in raw['fshunt']:
        """
        0,    1,      2,      3,      4
        Bus, name, Status, g (MW), b (Mvar)
        """
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        param = {
            'bus': bus,
            'Vn': vn,
            'u': data[2],
            'Sn': mva,
            'g': data[3] / mva,
            'b': data[4] / mva,
        }
        system.Shunt.elem_add(**param)

    gen_idx = 0
    for data in raw['gen']:
        """
         0, 1, 2, 3, 4, 5, 6, 7,    8,   9,10,11, 12, 13, 14,   15, 16,17,18,19
         I,ID,PG,QG,QT,QB,VS,IREG,MBASE,ZR,ZX,RT,XT,GTAP,STAT,RMPCT,PT,PB,O1,F1
        """
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        gen_mva = data[8]  # unused yet
        gen_idx += 1
        status = data[14]
        param = {
            'Sn': gen_mva,
            'Vn': vn,
            'u': status,
            'idx': gen_idx,
            'bus': bus,
            'pg': status * data[2] / mva,
            'qg': status * data[3] / mva,
            'qmax': data[4] / mva,
            'qmin': data[5] / mva,
            'v0': data[6],
            'ra': data[9],  # ra  armature resistance
            'xs': data[10],  # xs synchronous reactance
            'pmax': data[16] / mva,
            'pmin': data[17] / mva,
        }
        if data[0] in sw.keys():
            param.update({
                'a0': sw[data[0]],
            })
            system.SW.elem_add(**param)
        else:
            system.PV.elem_add(**param)

    for data in raw['branch']:
        """
        I,J,CKT,R,X,B,RATEA,RATEB,RATEC,GI,BI,GJ,BJ,ST,LEN,O1,F1,...,O4,F4
        """
        param = {
            'bus1': data[0],
            'bus2': data[1],
            'r': data[3],
            'x': data[4],
            'b': data[5],
            'rate_a': data[6],
            'Vn': system.Bus.get_field('Vn', data[0]),
            'Vn2': system.Bus.get_field('Vn', data[1]),
        }
        system.Line.elem_add(**param)

    xf_3_count = 1
    for data in raw['transf']:
        if len(data) == 4:
            """
            I,J,K,CKT,CW,CZ,CM,MAG1,MAG2,NMETR,'NAME',STAT,O1,F1,...,O4,F4
            R1-2,X1-2,SBASE1-2
            WINDV1,NOMV1,ANG1,RATA1,RATB1,RATC1,COD1,CONT1,RMA1,RMI1,VMA1,VMI1,NTP1,TAB1,CR1,CX1
            WINDV2,NOMV2
            """

            tap = data[2][0]
            phi = data[2][2]

            if tap == 1 and phi == 0:
                trasf = False
            else:
                trasf = True
            param = {
                'trasf': trasf,
                'bus1': data[0][0],
                'bus2': data[0][1],
                'u': data[0][11],
                'b': data[0][8],
                'r': data[1][0],
                'x': data[1][1],
                'tap': tap,
                'phi': phi,
                'rate_a': data[2][3],
                'Vn': system.Bus.get_field('Vn', data[0][0]),
                'Vn2': system.Bus.get_field('Vn', data[0][1]),
            }
            system.Line.elem_add(**param)
        else:
            """
            I, J, K, CKT, CW, CZ, CM, MAG1, MAG2, NMETR, 'NAME', STAT, Ol, Fl,...,o4, F4
            R1—2, X1—2, SBASE1—2, R2—3, X2—3, SBASE2—3, R3—1, X3—1, SBASE3—1, VMSTAR, ANSTAR
            WINDV1, NOMV1, ANG1, RATA1, BATB1, RATC1, COD1, CONT1, RMA1, RMI1, VMA1, VMI1, NTP1, TAB1, CR1, CX1
            WINDV2, NOMV2, ANG2, RATA2, BATB2, RATC2, COD2, CONT2, RMA2, RMI2, VMA2, VMI2, NTP2, TAB2, CR2, CX2
            WINDV3, NOMV3, ANG3, RATA3, BATB3, RATC3, COD3, CONT3, RMA3, RMI3, VMA3, VMI3, NTP3, TAB3, CR3, CX3
            """
            param = {
                'idx': max_bus+xf_3_count,
                'name': '_'.join(str(data[0][:3])),
                'Vn': system.Bus.get_field('Vn', data[0][0]),
                'voltage': data[1][-2],
                'angle': data[1][-1]/180*3.14
            }
            system.Bus.elem_add(**param)

            r = []
            x = []
            r.append((data[1][0] + data[1][6] - data[1][3])/2)
            r.append((data[1][3] + data[1][0] - data[1][6])/2)
            r.append((data[1][6] + data[1][3] - data[1][0])/2)
            x.append((data[1][1] + data[1][7] - data[1][4])/2)
            x.append((data[1][4] + data[1][1] - data[1][7])/2)
            x.append((data[1][7] + data[1][4] - data[1][1])/2)
            for i in range(0, 3):
                param = {
                    'trasf': True,
                    'bus1': data[0][i],
                    'bus2': max_bus+xf_3_count,
                    'u': data[0][11],
                    'b': data[0][8],
                    'r': r[i],
                    'x': x[i],
                    'tap': data[2+i][0],
                    'phi': data[2+i][2],
                    'rate_a': data[2+i][3],
                    'Vn': system.Bus.get_field('Vn', data[0][i]),
                    'Vn2': system.Bus.get_field('Vn', data[0][0]),
                }
                system.Line.elem_add(**param)
            xf_3_count += 1

    for data in raw['swshunt']:
        # I, MODSW, ADJM, STAT, VSWHI, VSWLO, SWREM, RMPCT, ’RMIDNT’,
        # BINIT, N1, B1, N2, B2, ... N8, B8
        bus = data[0]
        vn = system.Bus.get_field('Vn', bus)
        param = {
            'bus': bus,
            'Vn': vn,
            'Sn': mva,
            'u': data[3],
            'b': data[9] / mva,
        }
        system.Shunt.elem_add(**param)

    for data in raw['area']:
        """ID, ISW, PDES, PTOL, ARNAME"""
        param = {
            'idx': data[0],
            'isw': data[1],
            'pdes': data[2],
            'ptol': data[3],
            'name': data[4],
        }
        system.Area.elem_add(**param)

    for data in raw['zone']:
        """ID, NAME"""
        param = {
            'idx': data[0],
            'name': data[1],
        }
        system.Zone.elem_add(**param)

    return retval


def readadd(file, system):
    """read DYR file"""
    dyr = {}
    data = []
    end = 0
    retval = True
    sep = ','

    fid = open(file, 'r')
    for line in fid.readlines():
        if line.find('/') >= 0:
            line = line.split('/')[0]
            end = 1
        if line.find(',') >= 0:  # mixed comma and space splitter not allowed
            line = [to_number(item.strip()) for item in line.split(sep)]
        else:
            line = [to_number(item.strip()) for item in line.split()]
        if not line:
            end = 0
            continue
        data.extend(line)
        if end == 1:
            field = data[1]
            if field not in dyr.keys():
                dyr[field] = []
            dyr[field].append(data)
            end = 0
            data = []
    fid.close()

    # elem_add device elements to system
    supported = [
        'GENROU',
        'GENCLS',
        'ESST3A',
        'ESDC2A',
        'SEXS',
        'EXST1',
        'ST2CUT',
        'IEEEST',
        'TGOV1',
    ]
    used = list(supported)
    for model in supported:
        if model not in dyr.keys():
            used.remove(model)
            continue
        for data in dyr[model]:
            add_dyn(system, model, data)

    needed = list(dyr.keys())
    for i in supported:
        if i in needed:
            needed.remove(i)

    logger.warning('Models currently unsupported: {}'.format(
        ', '.join(needed)))

    return retval


def add_dyn(system, model, data):
    """helper function to elem_add a device element to system"""
    if model == 'GENCLS':
        bus = data[0]
        data = data[3:]
        if bus in system.PV.bus:
            dev = 'PV'
            gen_idx = system.PV.idx[system.PV.bus.index(bus)]
        elif bus in system.SW.bus:
            dev = 'SW'
            gen_idx = system.SW.idx[system.SW.bus.index(bus)]
        else:
            raise KeyError
        # todo: check xl
        idx_PV = get_idx(system, 'StaticGen', 'bus', bus)
        u = get_param(system, 'StaticGen', 'u', idx_PV)
        param = {
            'bus': bus,
            'gen': gen_idx,
            # 'idx': bus,
            #  use `bus` for `idx`. Only one generator allowed on each bus
            'Sn': system.__dict__[dev].get_field('Sn', gen_idx),
            'Vn': system.__dict__[dev].get_field('Vn', gen_idx),
            'xd1': system.__dict__[dev].get_field('xs', gen_idx),
            'ra': system.__dict__[dev].get_field('ra', gen_idx),
            'M': 2 * data[0],
            'D': data[1],
            'u': u,
        }
        system.Syn2.elem_add(**param)

    elif model == 'GENROU':
        bus = data[0]
        data = data[3:]
        if bus in system.PV.bus:
            dev = 'PV'
            gen_idx = system.PV.idx[system.PV.bus.index(bus)]
        elif bus in system.SW.bus:
            dev = 'SW'
            gen_idx = system.SW.idx[system.SW.bus.index(bus)]
        else:
            raise KeyError
        idx_PV = get_idx(system, 'StaticGen', 'bus', bus)
        u = get_param(system, 'StaticGen', 'u', idx_PV)

        param = {
            'bus': bus,
            'gen': gen_idx,
            # 'idx': bus,
            #  use `bus` for `idx`. Only one generator allowed on each bus
            'Sn': system.__dict__[dev].get_field('Sn', gen_idx),
            'Vn': system.__dict__[dev].get_field('Vn', gen_idx),
            'ra': system.__dict__[dev].get_field('ra', gen_idx),
            'Td10': data[0],
            'Td20': data[1],
            'Tq10': data[2],
            'Tq20': data[3],
            'M': 2 * data[4],
            'D': data[5],
            'xd': data[6],
            'xq': data[7],
            'xd1': data[8],
            'xq1': data[9],
            'xd2': data[10],
            'xq2': data[10],  # xd2 = xq2
            'xl': data[11],
            'u': u,
        }
        system.Syn6a.elem_add(**param)

    elif model == 'ESST3A':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[8],
            'vrmin': data[9],
            'Ka': data[6],
            'Ta': data[7],
            'Tf': data[5],
            'Tr': data[0],
            'Kf': data[5],
            'Ke': 1,
            'Te': 1,
        }
        system.AVR1.elem_add(**param)

    elif model == 'ESDC2A':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[5],
            'vrmin': data[6],
            'Ka': data[1],
            'Ta': data[2],
            'Tf': data[10],
            'Tr': data[0],
            'Kf': data[9],
            'Ke': 1,
            'Te': data[8],
        }
        system.AVR1.elem_add(**param)

    elif model == 'EXST1':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[7],
            'vrmin': data[8],
            'Ka': data[5],
            'Ta': data[6],
            'Kf': data[10],
            'Tf': data[11],
            'Tr': data[0],
            'Te': data[4],
        }
        system.AVR1.elem_add(**param)

    elif model == 'SEXS':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'syn': syn,
            'vrmax': data[5],
            'vrmin': data[4],
            'K0': data[2],
            'T2': data[1],
            'T1': data[0],
            'Te': data[3],
        }
        system.AVR3.elem_add(**param)

    elif model == 'IEEEG1':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)

        pass

    elif model == 'TGOV1':
        bus = data[0]
        data = data[3:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        param = {
            'gen': syn,
            'R': data[0],
            'T1': data[4],
            'T2': data[5],
        }
        system.TG2.elem_add(**param)

    elif model == 'ST2CUT':
        bus = data[0]
        data = data[3:]
        Ic1 = data[0]
        Ic2 = data[2]

        data = data[4:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        avr = get_idx(system, 'AVR', 'syn', syn)
        param = {
            'avr': avr,
            'Ic1': Ic1,
            'Ic2': Ic2,
            'K1': data[0],
            'K2': data[1],
            'T1': data[2],
            'T2': data[3],
            'T3': data[4],
            'T4': data[5],
            'T5': data[6],
            'T6': data[7],
            'T7': data[8],
            'T8': data[9],
            'T9': data[10],
            'T10': data[11],
            'lsmax': data[12],
            'lsmin': data[13],
            'vcu': data[14],
            'vcl': data[15],
        }
        system.PSS1.elem_add(**param)

    elif model == 'IEEEST':
        bus = data[0]
        data = data[3:]
        Ic = data[0]

        data = data[2:]
        syn = get_idx(system, 'Synchronous', 'bus', bus)
        avr = get_idx(system, 'AVR', 'syn', syn)
        param = {
            'avr': avr,
            'Ic': Ic,
            'A1': data[0],
            'A2': data[1],
            'A3': data[2],
            'A4': data[3],
            'A5': data[4],
            'A6': data[5],
            'T1': data[6],
            'T2': data[7],
            'T3': data[8],
            'T4': data[9],
            'T5': data[10],
            'T6': data[11],
            'Ks': data[12],
            'lsmax': data[13],
            'lsmin': data[14],
            'vcu': data[15],
            'vcl': data[16],
        }
        system.PSS2.elem_add(**param)

    else:
        logger.warning('Skipping unsupported model <{}> on bus {}'.format(
            model, data[0]))


def get_idx(system, group, param, fkey):
    ret = None
    for key, item in system.devman.group.items():
        if key != group:
            continue
        for name, dev in item.items():
            int_id = system.__dict__[dev].uid[name]
            if system.__dict__[dev].__dict__[param][int_id] == fkey:
                ret = name
                break
    return ret


def get_param(system, group, param, fkey):
    ret = None
    for key, item in system.devman.group.items():
        if key != group:
            continue
        for name, dev in item.items():
            if name == fkey:
                int_id = system.__dict__[dev].uid[name]
                ret = system.__dict__[dev].__dict__[param][int_id]
    return ret



import requests
import logging
import time  # NOQA
import os
import json
from andes.utils.math import to_number

logger = logging.getLogger(__name__)
sh = logging.StreamHandler()
logger.addHandler(sh)
logger.setLevel(logging.INFO)


class ANDESClient(object):

    def __init__(self, server_url):
        self.sysid = None
        self.server_url = server_url
        self.r = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unload_case()

    def load_case(self, case):
        """
        Use the REST API to load a data in andes server

        Parameters
        ----------
        case

        Returns
        -------

        """
        if self.sysid is not None:
            logger.error('sysid <{}> is not none. unload first.'.format(self.sysid))

        url = os.path.join(self.server_url, 'load')

        params = {'name': case}
        self.r = requests.get(url=url, params=params)
        logger.debug('Load case <{}> requested'.format(case))
        if not self.r.ok:
            logger.error('load <{}> error. file may not exist in server root.'.format(case))
            return False
        else:
            self.sysid = int(self.r.text)
            logger.info('laod_case returned sysid is <{}>'.format(self.sysid))
            return True

    def unload_case(self, force=False):
        """
        Unload the case previously loaded by this client

        Returns
        -------

        """
        url = os.path.join(self.server_url, 'unload')

        params = {'sysid': self.sysid, 'force': force}
        self.r = requests.get(url=url, params=params)
        logger.debug('Unload sysid <{}> requested'.format(self.sysid))

        if not self.r.ok:
            logger.error('Unload <{}> error. System might not have been loaded.'.format(self.sysid))
            return False
        else:
            self.sysid = None
            logger.info('Unload sent successful')
            return True

    def run_case(self, simulation_time):
        """
        Run a loaded system
        Parameters
        ----------
        simulation_time

        Returns
        -------

        """
        if simulation_time <= 0:
            logger.error('Simulation time must be greater than 0')
            return

        url = os.path.join(self.server_url, 'run')
        params = {'sysid': self.sysid, 'time': simulation_time}

        self.r = requests.get(url=url, params=params)
        logger.debug('Run case requested for <{}>s'.format(simulation_time))
        if not self.r.ok:
            logger.error('Run simulation error. System may not have been loaded.')
            return False
        else:
            ret = self.r.text.strip()
            logger.info('run_case returns value {}'.format(ret))
            return True

    def param_get(self, model, var_name=None, idx=None, sysbase=False):
        """
        Get parameter from the loaded system

        Parameters
        ----------
        server_url
        model
        var_name
        idx
        sysbase

        Returns
        -------

        """
        url = os.path.join(self.server_url, 'param')
        params = {'sysid': self.sysid, 'name': model, 'var': var_name, 'idx': idx, 'sysbase': sysbase}

        self.r = requests.get(url=url, params=params)

        logger.debug('param_get requested with {}'.format(params))

        if not self.r.ok:
            logger.error('Get parameter error. Check if the params are correct.')
            return False
        else:
            ret = to_number(self.r.text)
            logger.info('param_get returns value <{}>'.format(ret))
            return ret

    def param_set(self, model, var_name, idx, value, sysbase=False):
        """
        Set parameter for the loaded system

        Parameters
        ----------
        server_url
        model
        var_name
        idx
        value
        base

        Returns
        -------

        """
        url = os.path.join(self.server_url, 'param')
        params = {'sysid': self.sysid,
                  'name': model,
                  'var': var_name,
                  'idx': idx,
                  'value': value,
                  'sysbase': sysbase}

        self.r = requests.post(url=url, params=params)
        logger.debug('param_set requested with {}'.format(params))
        if not self.r.ok:
            logger.error('Set parameter error. Check if the params are correct.')
            return False
        else:
            ret = self.r.text.strip()
            logger.info('param_set returns value <{}>'.format(ret))
            return True

    def is_loaded(self):
        """
        Check if the server is loaded with a case

        Parameters
        ----------
        server_url

        Returns
        -------

        """
        url = os.path.join(self.server_url, 'status')
        params = {'sysid': self.sysid}

        self.r = requests.get(url=url, params=params)
        logger.debug('is_loaded requested for sysid <{}>'.format(self.sysid))
        status = int(json.loads(self.r.text))

        if status == 0:
            logger.info('is_loaded returned False')
            return False
        elif status == 1:
            logger.info('is_loaded returned True')
            return True
        else:
            logger.error('Unknown status {}'.format(status))


def fault_test(client):

    if client.param_get(model='Fault', var_name='u', idx=1) is not False:
        logger.info('Fault.1.u before: {}'.format(client.r.text))

    time.sleep(1)

    if client.param_set('Fault', 'u', '1', value=1, sysbase=True) is not False:
        logger.info('Fault.1.u after: {}'.format(client.r.text))


def avr_ka_test(client):
    if client.param_get(model='AVR1', var_name='Ka', idx=1) is not False:
        logger.info('AVR1.1.Ka before: {}'.format(client.r.text))

    time.sleep(5)
    if client.param_set(model='AVR1', var_name='Ka', idx=1, value=50, sysbase=False) is not False:
        pass

    if client.param_get(model='AVR1', var_name='Ka', idx=1) is not False:
        logger.info('AVR1.1.Ka after: {}'.format(client.r.text))


if __name__ == '__main__':

    local_server = 'http://127.0.0.1:5000'
    with ANDESClient(local_server) as client:
        r = client.load_case('ieee14_syn.dm')

        if client.is_loaded() is not False:
            client.run_case(20)

        fault_test(client)

        avr_ka_test(client)



from cvxopt import matrix, spmatrix  # NOQA
from cvxopt import mul, div, exp  # NOQA
from andes.consts import Fx, Fy, Gx, Gy, Fx0, Fy0, Gx0, Gy0  # NOQA
from .base import ModelBase
from andes.utils.math import zeros
import logging

logger = logging.getLogger(__name__)


class BArea(ModelBase):
    def __init__(self, system, name):
        super(BArea, self).__init__(system, name)
        self._group = 'Calculation'
        self._data.update({
            'area': None,
            'syn': None,
            'beta': 0,
        })
        self._mandatory.extend(['area', 'syn', 'beta'])
        self._algebs.extend(['Pexp', 'fcoi', 'ace'])
        self.calls.update({
            'gcall': True,
            'init1': True,
            'jac0': True,
        })
        self._service.extend(['P0', 'Mtot', 'M', 'usyn', 'wsyn'])
        self._fnamey.extend(['P_{exp}', 'f_{coi}', 'ace'])
        self._params.extend(['beta'])
        self._init()

    def init1(self, dae):

        for item in self._service:
            self.__dict__[item] = [[]] * self.n

        # Start with frequency
        for idx, item in enumerate(self.syn):
            self.M[idx] = self.read_data_ext(
                'Synchronous', field='M', idx=item)
            self.Mtot[idx] = sum(self.M[idx])
            self.usyn[idx] = self.read_data_ext(
                'Synchronous', field='u', idx=item)
            self.wsyn[idx] = self.read_data_ext(
                'Synchronous', field='omega', idx=item)
            dae.y[self.fcoi[idx]] = sum(
                mul(self.M[idx], dae.x[self.wsyn[idx]])) / self.Mtot[idx]

        # Get BA Export Power
        self.copy_data_ext('Area', field='area_P0', dest='P0', idx=self.area)
        dae.y[self.Pexp] = self.P0
        dae.y[self.ace] = 0

    def gcall(self, dae):

        P = self.read_data_ext('Area', field='area_P0', idx=self.area)
        dae.g[self.Pexp] = dae.y[self.Pexp] - P

        for idx, item in enumerate(self.syn):
            self.wsyn[idx] = self.read_data_ext(
                'Synchronous', field='omega', idx=item)
            dae.g[self.fcoi[idx]] = dae.y[self.fcoi[idx]] - sum(
                mul(self.M[idx], dae.x[self.wsyn[idx]])) / self.Mtot[idx]

        ACE = (P - self.P0) - mul(self.beta, (1 - dae.y[self.fcoi]))

        dae.g[self.ace] = dae.y[self.ace] + ACE

    def jac0(self, dae):
        dae.add_jac(Gy0, 1, self.Pexp, self.Pexp)
        dae.add_jac(Gy0, 1, self.fcoi, self.fcoi)
        dae.add_jac(Gy0, 1, self.ace, self.ace)
        dae.add_jac(Gy0, 1, self.ace, self.Pexp)
        dae.add_jac(Gy0, self.beta, self.ace, self.fcoi)


class AGC(ModelBase):
    def __init__(self, system, name):
        super(AGC, self).__init__(system, name)
        self._group = 'Control'
        self._data.update({
            'BArea': None,
            'syn': None,
            'Ki': 0.05,
        })
        self._mandatory.extend(['BArea', 'syn', 'Ki'])
        self._states.extend(['Pagc'])
        self.calls.update({
            'gcall': True,
            'init1': True,
            'jac0': True,
            'fcall': True,
        })
        self._service.extend(['ace', 'pm', 'M', 'usyn', 'Mtot'])
        self._fnamex.extend(['P_{agc tot}'])
        self._params.extend(['Ki'])
        self._init()

    def init1(self, dae):
        self.pm = [[]] * self.n
        self.M = [[]] * self.n
        self.usyn = [[]] * self.n
        self.Mtot = [[]] * self.n

        for idx, item in enumerate(self.syn):
            self.pm[idx] = self.read_data_ext(
                'Synchronous', field='pm', idx=item)
            self.usyn[idx] = self.read_data_ext(
                'Synchronous', field='u', idx=item)
            self.M[idx] = self.read_data_ext(
                'Synchronous', field='M', idx=item)
            self.Mtot[idx] = sum(mul(self.usyn[idx], self.M[idx]))

        self.copy_data_ext('BArea', field='ace', idx=self.BArea)

    def fcall(self, dae):
        dae.f[self.Pagc] = mul(self.Ki, dae.y[self.ace])

    def gcall(self, dae):
        for idx, item in enumerate(self.syn):
            Kgen = div(self.M[idx], self.Mtot[idx])
            dae.g[self.pm[idx]] -= mul(self.usyn[idx], Kgen,
                                       dae.x[self.Pagc[idx]])

    def jac0(self, dae):
        dae.add_jac(Fy0, self.Ki, self.Pagc, self.ace)

    def gycall(self, dae):
        for idx, item in enumerate(self.syn):
            Kgen = div(self.M[idx], self.Mtot[idx])
            dae.add_jac(Gx, -mul(self.usyn[idx], Kgen), self.pm[idx],
                        self.Pagc[idx])


class eAGC(ModelBase):
    def __init__(self, system, name):
        super(eAGC, self).__init__(system, name)
        self._group = 'Control'
        self._data.update({
            'cl': None,
            'tl': 0,
            'Pl': None,
            'BA': None,
        })
        self._descr.update({
            'cl': 'Loss sharing coefficient (vector)',
            'tl': 'Time of generator loss',
            'Pl': 'Loss of power generation in pu (vector)',
            'BA': 'Balancing Area that support the Gen loss',
        })
        self._mandatory.extend(['cl', 'tl', 'Pl', 'BA'])
        self.calls.update({
            'gcall': True,
            'init1': True,
            'jac0': False,
            'fcall': False,
        })
        self._service.extend(['ace', 'en'])
        self._params.extend(['cl', 'tl', 'Pl'])
        self._init()

    def init1(self, dae):
        self.ace = [[]] * self.n
        for idx, item in enumerate(self.BA):
            self.ace[idx] = self.read_data_ext('BArea', field='ace', idx=item)

        self.en = zeros(self.n, 1)

    def switch(self):
        """Switch if time for eAgc has come"""
        t = self.system.dae.t
        for idx in range(0, self.n):
            if t >= self.tl[idx]:
                if self.en[idx] == 0:
                    self.en[idx] = 1
                    logger.info(
                        'Extended ACE <{}> activated at t = {}.'.format(
                            self.idx[idx], t))

    def gcall(self, dae):
        self.switch()

        for idx in range(0, self.n):
            dae.g[self.ace[idx]] -= mul(self.en[idx], self.cl[:, idx],
                                        self.Pl[idx])



import os
import logging
import platform
import tempfile

logger = logging.getLogger(__name__)


def get_config_load_path(conf_path=None):
    """
    Return config file load path

    Priority:
        1. conf_path
        2. current directory
        3. home directory

    Parameters
    ----------
    conf_path

    Returns
    -------

    """

    if conf_path is None:
        # test ./andes.conf
        if os.path.isfile('andes.conf'):
            conf_path = 'andes.conf'
        # test ~/andes.conf
        home_dir = os.path.expanduser('~')
        if os.path.isfile(os.path.join(home_dir, '.andes', 'andes.conf')):
            conf_path = os.path.join(home_dir, '.andes', 'andes.conf')

    if conf_path is not None:
        logger.debug('Found config file at {}.'.format(conf_path))

    return conf_path


def get_log_dir():
    """
    Get a directory for logging

    On Linux or macOS, '/tmp/andes' is the default. On Windows,
    '%APPDATA%/andes' is the default.

    Returns
    -------
    str
        Path to the logging directory
    """
    PATH = ''
    if platform.system() in ('Linux', 'Darwin'):
        PATH = tempfile.mkdtemp(prefix='andes-')

    elif platform.system() == 'Windows':
        APPDATA = os.getenv('APPDATA')
        PATH = os.path.join(APPDATA, 'andes')

    if not os.path.exists(PATH):
        os.makedirs(PATH)
    return PATH



from cvxopt import matrix, spmatrix
from cvxopt import mul, div

from .base import ModelBase
from .measurement import BusFreq  # NOQA

from ..utils.math import aeqb, sdiv

from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA


class PSS1(ModelBase):
    """Stabilizer ST2CUT with dual input signals"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'PSS'
        self._name = 'PSS1'
        self._fnamey.extend([
            'In_1', 'In_2', 'In', 'x_3', 'x_4', 'x_5', 'x_6', 'V_{SS}',
            'V_{ST}'
        ])
        self._zeros.extend(['T4'])
        self._states.extend(['x1', 'x2', 'u3', 'u4', 'u5', 'u6'])
        self._params.extend([
            'Ic1', 'Ic2', 'K1', 'T1', 'K2', 'T2', 'T3', 'T4', 'T5', 'T6', 'T7',
            'T8', 'T9', 'T10', 'lsmax', 'lsmin', 'vcu', 'vcl', 'd1', 'd2', 'd3'
        ])
        self._times.extend(
            ['T1', 'T2', 'T3', 'T4', 'T5', 'T6', 'T7', 'T8', 'T9', 'T10'])
        self._mandatory.extend(['avr'])
        self._algebs.extend(
            ['In1', 'In2', 'In', 'x3', 'x4', 'x5', 'x6', 'vss', 'vst'])
        self._fnamex.extend(['x_1', 'x_2', 'u_3', 'u_4', 'u_5', 'u_6'])
        self._service.extend([
            'Ic23', 'Ic12', 'T34', 'toSg', 'Ic15', 'T910', 'Ic21', 'Ic14',
            'Ic13', 'Ic22', 'Ic24', 'T56', 'T78', 'v0', 'Ic11', 'Ic25', 'u0',
            'bus'
        ])
        self._units.update({
            'vcu': 'pu',
            'T4': 's',
            'T5': 's',
            'vcl': 'pu',
            'T10': 's',
            'lsmin': 'pu',
            'T9': 's',
            'T2': 's',
            'T7': 's',
            'T8': 's',
            'lsmax': 'pu',
            'T6': 's',
            'T1': 's',
            'T3': 's'
        })
        self._data.update({
            'K2': 0,
            'vcu': 0.25,
            'Ic2': 0,
            'T4': 1,
            'T5': 1.5,
            'lsmin': -0.2,
            'vcl': -0.25,
            'avr': 0,
            'T8': 0.15,
            'T9': 1.5,
            'T2': 0.02,
            'T7': 1.5,
            'K1': 0,
            'lsmax': 0.2,
            'Ic1': 0,
            'T10': 0.15,
            'T6': 0.15,
            'T1': 0.02,
            'T3': 1,
            'd1': 1,
            'd2': 1,
            'd3': 1
        })
        self._descr.update({
            'K2': 'Input 2 gain',
            'vcu': 'cutoff upper limit offset over Vt',
            'Ic2': 'Input 2 control switch',
            'T4': 'Washout time constant (denominator)',
            'T5': 'LL1 time constant (zero)',
            'lsmin': 'PSS output minimum limit',
            'vcl': 'cutoff lower limit offset over Vt',
            'Ic1': 'Input 1 control switch',
            'T8': 'LL2 time constant (pole)',
            'T9': 'LL3 time constant (zero)',
            'T2': 'Input 2 time constant',
            'T7': 'LL2 time constant (zero)',
            'K1': 'Input 1 gain',
            'lsmax': 'PSS output maximum limit',
            'T10': 'LL3 time constant (pole)',
            'avr': 'Exciter id',
            'T1': 'Input 1 time constant',
            'T3': 'Washout time constant (numerator)',
            'T6': 'LL1 time constant (pole)'
        })
        self.calls.update({
            'gcall': True,
            'fcall': True,
            'fxcall': False,
            'init1': True,
            'jac0': True,
            'gycall': False
        })
        self._zeros.extend(['T1', 'T2', 'T4'])
        self.param_remove('Vn')
        self.param_remove('Sn')
        self._init()
        # todo: fix T6, T8 or T10 == 10. Ignore the filter if they are zeros.

    def servcall(self, dae):
        self.copy_data_ext('AVR', 'syn', 'syn', self.avr)
        self.copy_data_ext('AVR', 'u', 'uavr', self.avr)
        self.copy_data_ext('Synchronous', 'u', 'usyn', self.syn)
        self.copy_data_ext('Synchronous', 'bus', 'bus', self.syn)
        self.copy_data_ext('Synchronous', 'Sn', 'Sg', self.syn)
        self.copy_data_ext('Synchronous', 'v', 'v', self.syn)
        self.copy_data_ext('Synchronous', 'vf', 'vf', self.syn)
        self.copy_data_ext('Synchronous', 'pm', 'pm', self.syn)
        self.copy_data_ext('Synchronous', 'omega', 'omega', self.syn)
        self.copy_data_ext('Synchronous', 'p', 'p', self.syn)
        self.copy_data_ext('BusFreq', 'w', 'w', self.bus)
        self.T34 = sdiv(self.T3, self.T4)
        self.T56 = sdiv(self.T5, self.T6)
        self.T78 = sdiv(self.T7, self.T8)
        self.T910 = sdiv(self.T9, self.T10)
        self.set_flag('T6', 'd1', reset_val=True)
        self.set_flag('T8', 'd2', reset_val=True)
        self.set_flag('T10', 'd3', reset_val=True)
        self.toSg = div(self.system.mva, self.Sg)
        self.v0 = dae.y[self.v]
        self.update_ctrl()

    def set_flag(self, value, flag, reset_val=False):
        """Set a flag to 0 if the corresponding value is 0"""
        if not self.__dict__[flag]:
            self.__dict__[flag] = matrix(1.0, (len(self.__dict__[value]), 1),
                                         'd')
        for idx, item in enumerate(self.__dict__[value]):
            if item == 0:
                self.__dict__[flag][idx] = 0
                if reset_val:
                    self.__dict__[value][idx] = 1

    def update_ctrl(self):
        self.u0 = mul(self.u, self.uavr,
                      self.usyn)  # effective PSS connectivity status
        self.Ic11 = aeqb(self.Ic1, 1)
        self.Ic12 = aeqb(self.Ic1, 2)
        self.Ic13 = aeqb(self.Ic1, 3)
        self.Ic14 = aeqb(self.Ic1, 4)
        self.Ic15 = aeqb(self.Ic1, 5)
        self.Ic21 = aeqb(self.Ic2, 1)
        self.Ic22 = aeqb(self.Ic2, 2)
        self.Ic23 = aeqb(self.Ic2, 3)
        self.Ic24 = aeqb(self.Ic2, 4)
        self.Ic25 = aeqb(self.Ic2, 5)
        # ignore the hard limiters if vcu == 0 or vcl == 0
        self.vcu += mul(aeqb(self.vcu, 0.0), 9999)
        self.vcl += mul(aeqb(self.vcl, 0.0), -9999)
        self.vtmax = self.v0 + self.vcu
        self.vtmin = self.v0 + self.vcl

    def init1(self, dae):
        self.servcall(dae)
        dae.y[self.In1] = mul(
            self.u0,
            mul(self.Ic11, -1 + dae.x[self.omega]) + mul(
                self.Ic12, -1 + dae.x[self.w]) + mul(self.Ic15, dae.y[self.v])
            + mul(self.Ic13, dae.y[self.p], self.toSg) + mul(
                self.Ic14, dae.y[self.pm], self.toSg))
        dae.y[self.In2] = mul(
            self.u0,
            mul(self.Ic21, -1 + dae.x[self.omega]) + mul(
                self.Ic22, -1 + dae.x[self.w]) + mul(self.Ic25, dae.y[self.v])
            + mul(self.Ic23, dae.y[self.p], self.toSg) + mul(
                self.Ic24, dae.y[self.pm], self.toSg))
        dae.x[self.x1] = mul(self.K1, dae.y[self.In1], self.u0)
        dae.x[self.x2] = mul(self.K2, dae.y[self.In2], self.u0)
        dae.y[self.In] = mul(self.u0, dae.y[self.In1] + dae.y[self.In2])
        dae.x[self.u3] = mul(dae.y[self.In], self.T34, self.u0)

    def gcall(self, dae):
        dae.g[self.In1] = mul(
            self.u0,
            -dae.y[self.In1] + mul(self.Ic11, -1 + dae.x[self.omega]) + mul(
                self.Ic12, -1 + dae.x[self.w]) + mul(self.Ic15, dae.y[self.v])
            + mul(self.Ic13, dae.y[self.p], self.toSg) + mul(
                self.Ic14, dae.y[self.pm], self.toSg))
        dae.g[self.In2] = mul(
            self.u0,
            -dae.y[self.In2] + mul(self.Ic21, -1 + dae.x[self.omega]) + mul(
                self.Ic22, -1 + dae.x[self.w]) + mul(self.Ic25, dae.y[self.v])
            + mul(self.Ic23, dae.y[self.p], self.toSg) + mul(
                self.Ic24, dae.y[self.pm], self.toSg))
        dae.g[self.In] = mul(
            self.u0, dae.y[self.In1] + dae.y[self.In2] - dae.y[self.In])
        dae.g[self.x3] = mul(
            self.u0,
            -dae.x[self.u3] - dae.y[self.x3] + mul(dae.y[self.In], self.T34))
        dae.g[self.x4] = mul(
            self.u0,
            dae.x[self.u4] - dae.y[self.x4] + mul(self.T56, dae.y[self.x3]))
        dae.g[self.x5] = mul(
            self.u0,
            dae.x[self.u5] - dae.y[self.x5] + mul(self.T78, dae.y[self.x4]))
        dae.g[self.x6] = mul(
            self.u0,
            dae.x[self.u6] - dae.y[self.x6] + mul(self.T910, dae.y[self.x5]))
        dae.g[self.vss] = mul(self.u0, dae.y[self.x6] - dae.y[self.vss])
        dae.hard_limit(self.vss, self.lsmin, self.lsmax)
        dae.g[self.vst] = mul(self.u0, dae.y[self.vss] - dae.y[self.vst])
        dae.hard_limit_remote(
            self.vst,
            self.v,
            rtype='y',
            rmin=self.vtmin,
            rmax=self.vtmax,
            min_yset=0,
            max_yset=0)
        dae.g += spmatrix(
            mul(self.u0, dae.y[self.vst]), self.vf, [0] * self.n, (dae.m, 1),
            'd')

    def fcall(self, dae):
        dae.f[self.x1] = mul(self.u0, div(1, self.T1),
                             -dae.x[self.x1] + mul(dae.y[self.In1], self.K1))
        dae.f[self.x2] = mul(self.u0, div(1, self.T2),
                             -dae.x[self.x2] + mul(dae.y[self.In2], self.K2))
        dae.f[self.u3] = mul(self.u0, div(1, self.T4),
                             -dae.x[self.u3] + mul(dae.y[self.In], self.T34))
        dae.f[self.u4] = mul(
            self.u0, self.d1, div(1, self.T6),
            -dae.x[self.u4] + mul(dae.y[self.x3], 1 - self.T56))
        dae.f[self.u5] = mul(
            self.u0, self.d2, div(1, self.T8),
            -dae.x[self.u5] + mul(dae.y[self.x4], 1 - self.T78))
        dae.f[self.u6] = mul(
            self.u0, self.d3, div(1, self.T10),
            -dae.x[self.u6] + mul(dae.y[self.x5], 1 - self.T910))

    def jac0(self, dae):
        dae.add_jac(Gy0, mul(self.Ic15, self.u0), self.In1, self.v)
        dae.add_jac(Gy0, -self.u0, self.In1, self.In1)
        dae.add_jac(Gy0, mul(self.Ic14, self.toSg, self.u0), self.In1, self.pm)
        dae.add_jac(Gy0, mul(self.Ic13, self.toSg, self.u0), self.In1, self.p)
        dae.add_jac(Gy0, mul(self.Ic25, self.u0), self.In2, self.v)
        dae.add_jac(Gy0, mul(self.Ic24, self.toSg, self.u0), self.In2, self.pm)
        dae.add_jac(Gy0, -self.u0, self.In2, self.In2)
        dae.add_jac(Gy0, mul(self.Ic23, self.toSg, self.u0), self.In2, self.p)
        dae.add_jac(Gy0, self.u0, self.In, self.In2)
        dae.add_jac(Gy0, self.u0, self.In, self.In1)
        dae.add_jac(Gy0, -self.u0, self.In, self.In)
        dae.add_jac(Gy0, -self.u0, self.x3, self.x3)
        dae.add_jac(Gy0, mul(self.T34, self.u0), self.x3, self.In)
        dae.add_jac(Gy0, mul(self.T56, self.u0), self.x4, self.x3)
        dae.add_jac(Gy0, -self.u0, self.x4, self.x4)
        dae.add_jac(Gy0, -self.u0, self.x5, self.x5)
        dae.add_jac(Gy0, mul(self.T78, self.u0), self.x5, self.x4)
        dae.add_jac(Gy0, -self.u0, self.x6, self.x6)
        dae.add_jac(Gy0, mul(self.T910, self.u0), self.x6, self.x5)
        dae.add_jac(Gy0, self.u0, self.vss, self.x6)
        dae.add_jac(Gy0, -self.u0, self.vss, self.vss)
        dae.add_jac(Gy0, self.u0, self.vst, self.vss)
        dae.add_jac(Gy0, -self.u0, self.vst, self.vst)
        dae.add_jac(Gy0, self.u0, self.vf, self.vst)
        dae.add_jac(Gx0, mul(self.Ic11, self.u0), self.In1, self.omega)
        dae.add_jac(Gx0, mul(self.Ic12, self.u0), self.In1, self.w)
        dae.add_jac(Gx0, mul(self.Ic21, self.u0), self.In2, self.omega)
        dae.add_jac(Gx0, mul(self.Ic22, self.u0), self.In2, self.w)
        dae.add_jac(Gx0, -self.u0, self.x3, self.u3)
        dae.add_jac(Gx0, self.u0, self.x4, self.u4)
        dae.add_jac(Gx0, self.u0, self.x5, self.u5)
        dae.add_jac(Gx0, self.u0, self.x6, self.u6)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T1)), self.x1, self.x1)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T2)), self.x2, self.x2)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T4)), self.u3, self.u3)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T6)), self.u4, self.u4)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T8)), self.u5, self.u5)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T10)), self.u6, self.u6)
        dae.add_jac(Fy0, mul(self.K1, self.u0, div(1, self.T1)), self.x1,
                    self.In1)
        dae.add_jac(Fy0, mul(self.K2, self.u0, div(1, self.T2)), self.x2,
                    self.In2)
        dae.add_jac(Fy0, mul(self.T34, self.u0, div(1, self.T4)), self.u3,
                    self.In)
        dae.add_jac(Fy0, mul(self.u0, div(1, self.T6), 1 - self.T56), self.u4,
                    self.x3)
        dae.add_jac(Fy0, mul(self.u0, div(1, self.T8), 1 - self.T78), self.u5,
                    self.x4)
        dae.add_jac(Fy0, mul(self.u0, div(1, self.T10), 1 - self.T910),
                    self.u6, self.x5)
        dae.add_jac(Gy0, 1e-6, self.In1, self.In1)
        dae.add_jac(Gy0, 1e-6, self.In2, self.In2)
        dae.add_jac(Gy0, 1e-6, self.In, self.In)
        dae.add_jac(Gy0, 1e-6, self.x3, self.x3)
        dae.add_jac(Gy0, 1e-6, self.x4, self.x4)
        dae.add_jac(Gy0, 1e-6, self.x5, self.x5)
        dae.add_jac(Gy0, 1e-6, self.x6, self.x6)
        dae.add_jac(Gy0, 1e-6, self.vss, self.vss)
        dae.add_jac(Gy0, 1e-6, self.vst, self.vst)


class PSS2(ModelBase):
    """Stabilizer IEEEST"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'PSS'
        self._name = 'PSS2'
        self._algebs.extend(['In', 'v1', 'v2', 'v3', 'v4', 'vss', 'vst'])
        self._fnamex.extend(['q_0', 'q_1', 'q_2', 'q_3', 'x_1', 'x_2', 'x_3'])
        self._fnamey.extend(
            ['In', 'v_1', 'v_2', 'v_3', 'v_4', 'V_{SS}', 'V_{ST}'])
        self._mandatory.extend(['avr'])
        self._params.extend([
            'A1', 'A2', 'A3', 'A4', 'A5', 'A6', 'Ks', 'T1', 'T2', 'T3', 'T4',
            'T5', 'T6', 'lsmax', 'lsmin', 'vcu', 'vcl', 'Ic'
        ])
        self._service.extend([
            'Ic3', 'Ic5', 'Ic1', 'H2', 'H1', 'H3', 'T34', 'T56', 'T12', 'toSg',
            'Ic2', 'Ic6', 'Ic4', 'u0', 'KsT56', 'H4'
        ])
        self._states.extend(['q0', 'q1', 'q2', 'q3', 'x1', 'x2', 'x3'])
        self._times.extend(['T1', 'T2', 'T3', 'T4', 'T5', 'T6'])
        self._data.update({
            'A4': 1,
            'lsmax': 0.1,
            'u': 1,
            'A1': 1,
            'Ks': -2,
            'lsmin': -0.1,
            'avr': 0,
            'T4': 0.75,
            'T6': 0.2,
            'T2': 0.02,
            'A5': 1,
            'T5': 1,
            'T1': 0.02,
            'A2': 1,
            'A6': 1,
            'T3': 1,
            'Ic': 0,
            'A3': 1,
            'vcu': 1.2,
            'vcl': 0.8
        })
        self._descr.update({
            'A4':
            'Filter-1 denominator coefficient',
            'lsmax':
            'PSS output maximum limit',
            'A6':
            'Filter-1 numerator coefficient',
            'A1':
            'Filter-1 denominator coefficient (1 + s*A1 + s^2 * A2)',
            'Ks':
            'Gain',
            'lsmin':
            'PSS output minimum limit',
            'avr':
            'Exciter id',
            'T4':
            'Filter-3 denominator time constant',
            'T6':
            'Washout denominator time constant',
            'A5':
            'Filter-1 numerator coefficient (1 + s*A5 + s^2 * A6)',
            'T5':
            'Washout numerator time constant',
            'T1':
            'Filter-2 numerator time constant',
            'A2':
            'Filter-1 denominator coefficient',
            'T2':
            'Filter-2 denominator time constant',
            'T3':
            'Filter-3 numerator time constant',
            'Ic':
            'Input signal selector (1) shaft speed dev (2) bus freq dev '
            '(3) P in syn base (4) Pm in syn base (5) Vbus',
            'A3':
            'Filter-1 denominator coefficient (1 + s*A3 + s^2 * A4)',
            'vcu':
            'Cutoff upper limit',
            'vcl':
            'Cutoff lower limit'
        })
        self._units.update({
            'A4': 'pu',
            'lsmax': 'pu',
            'A6': 'pu',
            'A1': 'pu',
            'Ks': 'pu',
            'lsmin': 'pu',
            'vcl': 'pu',
            'T4': 's',
            'T6': 's',
            'A5': 'pu',
            'T5': 's',
            'T1': 's',
            'A2': 'pu',
            'T2': 's',
            'T3': 's',
            'A3': 'pu',
            'vcu': 'pu'
        })
        self._zeros.extend(['T2', 'T4', 'T6', 'A2', 'A4'])
        self.calls.update({
            'fcall': True,
            'fxcall': False,
            'jac0': True,
            'init1': True,
            'gycall': False,
            'gcall': True
        })
        self.param_remove('Vn')
        self.param_remove('Sn')
        self._init()

    def servcall(self, dae):
        self.copy_data_ext('AVR', 'syn', 'syn', self.avr)
        self.copy_data_ext('AVR', 'u', 'uavr', self.avr)
        self.copy_data_ext('Synchronous', 'bus', 'bus', self.syn)
        self.copy_data_ext('Synchronous', 'u', 'usyn', self.syn)
        self.copy_data_ext('Synchronous', 'Sn', 'Sg', self.syn)
        self.copy_data_ext('Synchronous', 'omega', 'omega', self.syn)
        self.copy_data_ext('Synchronous', 'v', 'v', self.syn)
        self.copy_data_ext('Synchronous', 'p', 'p', self.syn)
        self.copy_data_ext('Synchronous', 'pm', 'pm', self.syn)
        self.copy_data_ext('Synchronous', 'vf', 'vf', self.syn)
        self.copy_data_ext('BusFreq', 'w', 'w', self.bus)
        self.u0 = mul(self.u, self.uavr, self.usyn)
        self.H1 = self.A1 + self.A3
        self.H2 = self.A2 + mul(self.A1, self.A3, self.A4)
        self.H3 = mul(self.A1, self.A4) + mul(self.A2, self.A3)
        self.H4 = mul(self.A2, self.A4)
        self.T12 = mul(self.T1, div(1, self.T2))
        self.T34 = mul(self.T3, div(1, self.T4))
        self.T56 = mul(self.T5, div(1, self.T6))
        self.KsT56 = mul(self.Ks, self.T5, div(1, self.T6))
        self.toSg = self.system.mva * div(1, self.Sg)
        self.update_ctrl()

    def update_ctrl(self):
        self.Ic1 = aeqb(self.Ic, 1)
        self.Ic2 = aeqb(self.Ic, 2)
        self.Ic3 = aeqb(self.Ic, 3)
        self.Ic4 = aeqb(self.Ic, 4)
        self.Ic5 = aeqb(self.Ic, 5)
        self.lsmax += mul(aeqb(self.lsmax, 0.0), 9999)
        self.lsmin += mul(aeqb(self.lsmin, 0.0), -9999)

    def init1(self, dae):
        self.servcall(dae)
        dae.y[self.In] = mul(
            self.u0,
            mul(self.Ic1, -1 + dae.x[self.omega]) + mul(
                self.Ic2, -1 + dae.x[self.w]) + mul(self.Ic5, dae.y[self.v]) +
            mul(self.Ic3, dae.y[self.p], self.toSg) + mul(
                self.Ic4, dae.y[self.pm], self.toSg))
        dae.x[self.q0] = dae.y[self.In]
        dae.y[self.v1] = dae.y[self.In]
        dae.x[self.x1] = mul(dae.y[self.v1], 1 - self.T12)
        dae.y[self.v2] = dae.y[self.v1]
        dae.x[self.x2] = mul(dae.y[self.v2], 1 - self.T34)
        dae.y[self.v3] = dae.y[self.v2]
        dae.x[self.x3] = mul(self.KsT56, dae.y[self.v3])

    def gcall(self, dae):
        dae.g[self.In] = mul(
            self.u0,
            -dae.y[self.In] + mul(self.Ic1, -1 + dae.x[self.omega]) + mul(
                self.Ic2, -1 + dae.x[self.w]) + mul(self.Ic5, dae.y[self.v]) +
            mul(self.Ic3, dae.y[self.p], self.toSg) + mul(
                self.Ic4, dae.y[self.pm], self.toSg))
        dae.g[self.v1] = mul(
            self.u0, dae.x[self.q0] - dae.y[self.v1] + mul(
                self.A5, dae.x[self.q1]) + mul(self.A6, dae.x[self.q2]))
        dae.g[self.v2] = mul(
            self.u0,
            dae.x[self.x1] - dae.y[self.v2] + mul(self.T12, dae.y[self.v1]))
        dae.g[self.v3] = mul(
            self.u0,
            dae.x[self.x2] - dae.y[self.v3] + mul(self.T34, dae.y[self.v2]))
        dae.g[self.v4] = mul(
            self.u0,
            -dae.y[self.v4] - dae.x[self.x3] + mul(self.KsT56, dae.y[self.v3]))
        dae.g[self.vss] = mul(self.u0, dae.y[self.v4] - dae.y[self.vss])
        dae.hard_limit(self.vss, self.lsmin, self.lsmax)
        dae.g[self.vst] = mul(self.u0, dae.y[self.vss] - dae.y[self.vst])
        dae.g += spmatrix(
            mul(self.u0, dae.y[self.vst]), self.vf, [0] * self.n, (dae.m, 1),
            'd')

    def fcall(self, dae):
        dae.f[self.q0] = mul(dae.x[self.q1], self.u0)
        dae.f[self.q1] = mul(dae.x[self.q2], self.u0)
        dae.f[self.q2] = mul(dae.x[self.q3], self.u0)
        dae.f[self.q3] = mul(
            self.u0, div(1, self.H4),
            dae.y[self.In] - dae.x[self.q0] - mul(self.H1, dae.x[self.q1]) -
            mul(self.H2, dae.x[self.q2]) - mul(self.H3, dae.x[self.q3]))
        dae.f[self.x1] = mul(
            self.u0, div(1, self.T2),
            -dae.x[self.x1] + mul(dae.y[self.v1], 1 - self.T12))
        dae.f[self.x2] = mul(
            self.u0, div(1, self.T4),
            -dae.x[self.x2] + mul(dae.y[self.v2], 1 - self.T34))
        dae.f[self.x3] = mul(self.u0, div(1, self.T6),
                             -dae.x[self.x3] + mul(self.KsT56, dae.y[self.v3]))

    def jac0(self, dae):
        dae.add_jac(Gy0, mul(self.Ic5, self.u0), self.In, self.v)
        dae.add_jac(Gy0, -self.u0, self.In, self.In)
        dae.add_jac(Gy0, mul(self.Ic4, self.toSg, self.u0), self.In, self.pm)
        dae.add_jac(Gy0, mul(self.Ic3, self.toSg, self.u0), self.In, self.p)
        dae.add_jac(Gy0, -self.u0, self.v1, self.v1)
        dae.add_jac(Gy0, mul(self.T12, self.u0), self.v2, self.v1)
        dae.add_jac(Gy0, -self.u0, self.v2, self.v2)
        dae.add_jac(Gy0, -self.u0, self.v3, self.v3)
        dae.add_jac(Gy0, mul(self.T34, self.u0), self.v3, self.v2)
        dae.add_jac(Gy0, -self.u0, self.v4, self.v4)
        dae.add_jac(Gy0, mul(self.KsT56, self.u0), self.v4, self.v3)
        dae.add_jac(Gy0, self.u0, self.vss, self.v4)
        dae.add_jac(Gy0, -self.u0, self.vss, self.vss)
        dae.add_jac(Gy0, -self.u0, self.vst, self.vst)
        dae.add_jac(Gy0, self.u0, self.vst, self.vss)
        dae.add_jac(Gy0, self.u0, self.vf, self.vst)
        dae.add_jac(Gx0, mul(self.Ic2, self.u0), self.In, self.w)
        dae.add_jac(Gx0, mul(self.Ic1, self.u0), self.In, self.omega)
        dae.add_jac(Gx0, mul(self.A5, self.u0), self.v1, self.q1)
        dae.add_jac(Gx0, self.u0, self.v1, self.q0)
        dae.add_jac(Gx0, mul(self.A6, self.u0), self.v1, self.q2)
        dae.add_jac(Gx0, self.u0, self.v2, self.x1)
        dae.add_jac(Gx0, self.u0, self.v3, self.x2)
        dae.add_jac(Gx0, -self.u0, self.v4, self.x3)
        dae.add_jac(Fx0, self.u0, self.q0, self.q1)
        dae.add_jac(Fx0, self.u0, self.q1, self.q2)
        dae.add_jac(Fx0, self.u0, self.q2, self.q3)
        dae.add_jac(Fx0, -mul(self.H3, self.u0, div(1, self.H4)), self.q3,
                    self.q3)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.H4)), self.q3, self.q0)
        dae.add_jac(Fx0, -mul(self.H2, self.u0, div(1, self.H4)), self.q3,
                    self.q2)
        dae.add_jac(Fx0, -mul(self.H1, self.u0, div(1, self.H4)), self.q3,
                    self.q1)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T2)), self.x1, self.x1)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T4)), self.x2, self.x2)
        dae.add_jac(Fx0, -mul(self.u0, div(1, self.T6)), self.x3, self.x3)
        dae.add_jac(Fy0, mul(self.u0, div(1, self.H4)), self.q3, self.In)
        dae.add_jac(Fy0, mul(self.u0, div(1, self.T2), 1 - self.T12), self.x1,
                    self.v1)
        dae.add_jac(Fy0, mul(self.u0, div(1, self.T4), 1 - self.T34), self.x2,
                    self.v2)
        dae.add_jac(Fy0, mul(self.KsT56, self.u0, div(1, self.T6)), self.x3,
                    self.v3)
        dae.add_jac(Gy0, 1e-6, self.In, self.In)
        dae.add_jac(Gy0, 1e-6, self.v1, self.v1)
        dae.add_jac(Gy0, 1e-6, self.v2, self.v2)
        dae.add_jac(Gy0, 1e-6, self.v3, self.v3)
        dae.add_jac(Gy0, 1e-6, self.v4, self.v4)
        dae.add_jac(Gy0, 1e-6, self.vss, self.vss)
        dae.add_jac(Gy0, 1e-6, self.vst, self.vst)



from cvxopt import matrix, spmatrix  # NOQA
from cvxopt import mul, div, exp  # NOQA
from ..consts import *  # NOQA
from ..utils.math import neg
from .base import ModelBase
import logging

logger = logging.getLogger(__name__)


class Breaker(ModelBase):
    """Simple line breaker model"""

    def __init__(self, system, name):
        super(Breaker, self).__init__(system, name)
        self._group = 'Relay'
        self._name = 'Breaker'
        self._data.update({
            'bus': None,
            'line': None,
            't1': 0.0,
            't2': 0.0,
            't3': 0.0,
            't4': 0.0,
            'u1': False,
            'u2': False,
            'u3': False,
            'u4': False,
            'fn': 60.0,
        })
        self._params.extend(['t1', 't2', 't3', 't4', 'u1', 'u2', 'u3', 'u4'])
        self._descr.update({
            'bus': 'Bus idx',
            'line': 'Line idx',
            't1': 'Time of the 1st switch',
            't2': 'Time of the 2nd switch',
            't3': 'Time of the 3rd switch',
            't4': 'Time of the 4th switch',
            'u1': 'Apply the 1st switch',
            'u2': 'Apply the 2nd switch',
            'u3': 'Apply the 3rd switch',
            'u4': 'Apply the 4th switch',
            'fn': 'rated frequency',
        })
        self._units.update({
            't1': 's',
            't2': 's',
            't3': 's',
            't4': 's',
            'u1': 'boolean',
            'u2': 'boolean',
            'u3': 'boolean',
            'u4': 'boolean',
        })
        self._mandatory.extend(['bus', 'line'])
        self._service.extend(['times', 'time'])
        self.param_remove('Sn')
        self._init()

    def setup(self):
        super(Breaker, self).setup()

        self.copy_data_ext('Line', 'bus1', dest='bus1', idx=self.line)
        self.copy_data_ext('Line', 'bus2', dest='bus2', idx=self.line)

        # check if `self.bus` is connected to `self.line`
        for i in range(self.n):
            if self.bus[i] not in (self.bus1[i], self.bus2[i]):
                logger.warning(
                    '<Breaker> {} on line {} and bus {} is incorrect '
                    'and is thus disabled.'.
                    format(self.idx[i], self.line[i], self.bus[i]))
                self.u[i] = 0

    def get_times(self):
        """Return all the action times and times-1e-6 in a list"""
        if not self.n:
            return []
        self.times = list(mul(self.u1, self.t1)) + \
            list(mul(self.u2, self.t2)) + \
            list(mul(self.u3, self.t3)) + \
            list(mul(self.u4, self.t4))

        self.times = matrix(list(set(self.times)))
        self.times = list(self.times) + list(self.times - 1e-6)
        return self.times

    def is_time(self, t):
        if not self.n:
            return
        return t in self.times

    def apply(self, actual_time):
        if self.time != actual_time:
            self.time = actual_time
        else:
            return

        n_applied = 0
        for i in range(self.n):
            tn = matrix([self.t1[i], self.t2[i], self.t3[i], self.t4[i]])
            tn = mul(self.u[i], tn)
            if actual_time in tn:

                line_int = self.system.Line.uid[self.line[i]]
                u0 = self.system.Line.u[line_int]
                self.system.Line.switch(self.line[i], neg(u0))

                if u0 == 1:
                    inf = ' Breaker <{}>: Line <{}> disconnected ' \
                          'at t = {}.'.format(
                              self.idx[i], self.line[i], actual_time)
                    n_applied += 1
                elif u0 == 0:
                    inf = ' Breaker <{}>: Line <{}> reconnected ' \
                          'at t = {}.'.format(
                              self.idx[i], self.line[i], actual_time)
                    n_applied += 1
                logger.info(inf)
        if n_applied > 0:
            self.system.check_islands(show_info=True)

    def insert(self, idx=None, name=None, **kwargs):
        if self.n:
            self._param_to_list()

        self.elem_add(idx, name, **kwargs)
        self._param_to_matrix()



from cvxopt import matrix, spdiag, mul, div, spmatrix, sparse  # NOQA

from .base import ModelBase

from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA


class SeriesBase(ModelBase):
    """Base class for AC series devices"""
    def __init__(self, system, name):
        super(SeriesBase, self).__init__(system, name)



from cvxopt import matrix, mul

from .base import ModelBase

from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA

from ..utils.math import sort_idx

import logging
logger = logging.getLogger(__name__)


class Stagen(ModelBase):
    """Static generator base class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'StaticGen'
        self._data.update({
            'bus': None,
            'busr': None,
            'pg': 0.0,
            'qg': 0.0,
            'pmax': 1.0,
            'pmin': 0.0,
            'qmax': 0.0,
            'qmin': 0.0,
            'v0': 1.0,
            'vmax': 1.4,
            'vmin': 0.6,
            'ra': 0.01,
            'xs': 0.3,
        })
        self._units.update({
            'bus': 'na',
            'busr': 'na',
            'pg': 'pu',
            'qg': 'pu',
            'pmax': 'pu',
            'pmin': 'pu',
            'qmax': 'pu',
            'v0': 'pu',
            'vmax': 'pu',
            'vmin': 'pu',
        })
        self._params.extend([
            'v0',
            'pg',
            'qg',
            'pmax',
            'pmin',
            'qmax',
            'qmin',
            'vmax',
            'vmin',
            'ra',
            'xs',
        ])
        self._descr.update({
            'bus': 'the idx of the installed bus',
            'busr': 'the idx of remotely controlled bus',
            'pg': 'active power set point',
            'qg': 'reactive power set point',
            'pmax': 'maximum active power output',
            'pmin': 'minimum active power output',
            'qmax': 'maximim reactive power output',
            'qmin': 'minimum reactive power output',
            'v0': 'voltage set point',
            'vmax': 'maximum voltage voltage',
            'vmin': 'minimum allowed voltage',
        })
        self._ac = {'bus': ['a', 'v']}
        # self._powers = ['pg', 'qg', 'pmax', 'pmin', 'qmax', 'qmin']
        self._voltages = ['v0', 'vmax', 'vmin']
        self._service = []
        self.calls.update({
            'gcall': True,
            'gycall': True,
            'init0': True,
            'pflow': True,
            'jac0': True,
            'stagen': True,
        })


class PV(Stagen):
    """Static PV generator for power flow"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'PV'
        self._algebs.extend(['q'])
        self._unamey = ['Q']
        self._fnamey = ['Q']
        self._service.extend(['qlim', 'above', 'below'])
        self._init()

    def init0(self, dae):
        """
        Set initial voltage and reactive power for PQ.
        Overwrites Bus.voltage values
        """
        dae.y[self.v] = self.v0
        dae.y[self.q] = mul(self.u, self.qg)

    def gcall(self, dae):
        update_qlim = False
        if self.system.pflow.config.pv2pq:
            pflow = self.system.pflow
            if pflow.niter >= pflow.config.ipv2pq:
                update_qlim = True
            elif len(pflow.iter_mis) > 0 and pflow.iter_mis[-1] <= min(0.01, 1e4 * pflow.config.tol):
                update_qlim = True

        qlim_new = []
        if update_qlim is True:
            d_min = dae.y[self.q] - self.qmin
            d_max = dae.y[self.q] - self.qmax
            idx_asc = sort_idx(d_min)
            idx_desc = sort_idx(d_max, reverse=True)

            nabove = nbelow = int(self.system.pflow.config.npv2pq)
            nconv = int(min(self.system.pflow.config.npv2pq, self.n))

            for i in range(nconv - 1, -1, -1):
                if d_min[idx_asc[i]] >= 0:
                    nbelow -= 1
                if d_max[idx_desc[i]] <= 0:
                    nabove -= 1

            self.below = idx_asc[0:nbelow] if nbelow else []
            self.above = idx_desc[0:nabove] if nabove else []
            mq = matrix(self.q)
            dae.y[mq[self.below]] = self.qmin[self.below]
            dae.y[mq[self.above]] = self.qmax[self.above]
            qlim_new = list(set(list(mq[self.below]) + list(mq[self.above])))

            if qlim_new:
                # refactorize the DAE when limit is hit. It allow resetting the Jacobian elements
                self.system.dae.factorize = True

        self.qlim = list(set(list(self.qlim) + qlim_new))

        p_inj = -mul(self.u, self.pg)
        q_inj = -mul(self.u, dae.y[self.q])
        v_mis = mul(self.u, dae.y[self.v] - self.v0)

        # TODO: improve readability
        for a, v, q, pi, qi, dv in zip(self.a, self.v, self.q, p_inj, q_inj,
                                       v_mis):
            dae.g[a] += pi
            dae.g[v] += qi
            dae.g[q] = dv  # not an interface equation

        if self.qlim:
            dae.g[self.qlim] = 0

    def gycall(self, dae):
        for q in self.qlim:
            v = self.v[self.q.index(q)]
            dae.set_jac('Gy', 0.0, q, v)
            dae.set_jac('Gy', 0.0, v, q)
            dae.set_jac('Gy', 1.0, q, q)

    def jac0(self, dae):
        dae.set_jac('Gy0', -1e-6, self.v, self.v)
        dae.set_jac('Gy0', -self.u, self.v, self.q)
        dae.set_jac('Gy0', self.u, self.q, self.v)
        dae.set_jac('Gy0', -1e-6, self.q, self.q)

    def disable_gen(self, idx):
        """
        Disable a PV element for TDS

        Parameters
        ----------
        idx

        Returns
        -------

        """
        self.u[self.uid[idx]] = 0
        self.system.dae.factorize = True


class Slack(PV):
    """Static slack generator"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'SW'
        self._data.update({'a0': 0.0})
        self._params.extend(['a0'])
        self._units.update({'a0': 'rad'})
        self._descr.update({'a0': 'reference angle'})
        self._algebs.extend(['p'])
        self._unamey.extend(['P'])
        self._fnamey.extend(['P'])
        # self._service.extend(['a0'])
        self.calls.update({'gycall': False})
        self._init()

    def init0(self, dae):
        super().init0(dae)
        self.a0 = self.system.Bus.angle[self.a]
        dae.y[self.p] = mul(self.u, self.pg)

    def gcall(self, dae):
        p_inj = -mul(self.u, dae.y[self.p])
        q_inj = -mul(self.u, dae.y[self.q])
        v_mis = mul(self.u, dae.y[self.v] - self.v0)
        a_mis = mul(self.u, dae.y[self.a] - self.a0)
        for a, v, p, q, pi, qi, dv, da in zip(self.a, self.v, self.p, self.q,
                                              p_inj, q_inj, v_mis, a_mis):
            dae.g[a] += pi
            dae.g[v] += qi
            dae.g[q] = dv
            dae.g[p] = da

        # dae.g[self.a] -= mul(self.u, dae.y[self.p])
        # dae.g[self.v] -= mul(self.u, dae.y[self.q])
        # dae.g[self.q] = mul(self.u, dae.y[self.v] - self.v0)
        # dae.g[self.p] = mul(self.u, dae.y[self.a] - self.a0)

    def jac0(self, dae):
        super().jac0(dae)
        dae.set_jac('Gy0', -self.u, self.a, self.p)
        dae.set_jac('Gy0', self.u, self.p, self.a)
        dae.set_jac('Gy0', 1e-6, self.p, self.p)



from .devman import DevMan  # NOQA
from .fileman import FileMan  # NOQA
from .dae import DAE  # NOQA
from .varname import VarName  # NOQA
from .varout import VarOut  # NOQA
from .call import Call  # NOQA
from .report import Report  # NOQA
